"""SynthDesc.sc"""

import pathlib
import json
import io

from ..base import platform as plf
from ..base import utils as utl
from ..base import systemactions as sac
from ..base import classlibrary as clb
from ..base import model as mdl
from ..base import main as _libsc3
from . import spec as spc
from . import ugen as ugn
from . import ugens as ugns
from . import server as srv
from . import synthdef as sdf
from . import _fmtrw as frw
from .ugens import inout as iou


__all__ = ['SynthDescLib']


class MdPlugin():
    # Metadata is just a JSON object represented as a Python dictionary.
    # The parameter ``codec`` is a dict containing 'key codecs' ordered by key.
    # Key codecs are objects of three fields, key, encoder and decoder,
    # they know to encode non serializable Python objects and convert them
    # back for a particular metadata key (all the content of a key, e.g.
    # spec_codec knows to decode specs which is the only data type of the key),
    # encoder is function that return JSON serializable Python objects,
    # it acts just like storeArgs in sclang. decoder receives that objects
    # retrieved with json.load an reconstruct the right object for the key.
    # Custom MdPlugins are instance objects with a dict of key codecs, there
    # is no need to make subclasses. This is so we don't have to write a
    # serialization protocol for each sc3 object. Different metadata
    # organizations can be differentiated just by its content (e.g. special
    # keys, format keys, version keys or similar).

    SUFFIX = 'scjsonmd'  # The only extension.
    _default_codec = {'specs': spc.spec_codec}

    def __init__(self, codec=None):
        self.codec = codec or type(self)._default_codec

    def write(self, synthdef, path):  # Was *write_metadata.
        path = pathlib.Path(path) / f'{synthdef.name}.{self.SUFFIX}'
        if path.exists():
            path.unlink()
        if synthdef.metadata:
            metadata = synthdef.metadata.copy()
            for key in self.codec:
                if key in metadata:
                    metadata[key] = self.codec[key].encoder(metadata[key])
            with open(path, 'w') as file:
                json.dump(metadata, file)

    def read(self, synthdef, path):  # Was *read_metadata.
        path = pathlib.Path(path) / f'{synthdef.name}.{self.SUFFIX}'
        return self.read_file(path)

    def read_file(self, path):
        if path.exists():
            with open(path, 'r') as file:
                metadata = json.load(file)
            for key in self.codec:
                if key in metadata:
                    metadata[key] = self.codec[key].decoder(metadata[key])
            return metadata
        return None

    def delete(self, synthdef, path):  # Was *clear_metadata.
        path = pathlib.Path(path) / f'{synthdef.name}.{self.SUFFIX}'
        if path.exists():
            path.unlink()

    # def delete_all(self, path):
    #     for file in pahtlib.Path(path).glob(f'*.{self.SUFFIX}'):
    #         file.unlink()


class SynthDescError(Exception):
    pass


class IODesc():
    def __init__(self, rate, num_channels, starting_channel, type):
        self.rate = rate
        self.num_channels = num_channels
        self.starting_channel = starting_channel or '?'
        self.type = type

    def __repr__(self):  # Was printOn.
        return (
            f"{type(self).__name__}(rate='{self.rate}', "
            f"num_channels={self.num_channels}, "
            f"starting_channel={self.starting_channel}, "
            f"type={self.type.__name__})")


class SynthDesc():
    _RATE_NAME = ('scalar', 'control', 'audio', 'demand')  # Used by index.

    md_plugin = MdPlugin()  # TODO: // override in your startup file.
    populate_metadata_func = lambda desc: None

    def __init__(self):
        self.name = None
        self.control_names = None
        self.control_dict = None
        self.controls = None
        self.inputs = None
        self.outputs = None
        self.metadata = None

        self.constants = None
        self.sdef = None  # Was def, reserved word.
        self.has_gate = False
        self.keep_gate = False  # Was msg_func_keep_gate.
        self.has_array_args = None
        self.has_variants = False
        # self.can_free_synth = False  # Non core interface, removed.

    @classmethod
    def new_from(cls, synthdef, keep_def=True):
        stream = io.BytesIO(synthdef.as_bytes())
        stream.read(4)  # SCgf
        version = frw.read_i32(stream)
        num_defs = frw.read_i16(stream)  # Always 1 here. Not used.
        desc = cls()  # desc = SynthDesc()
        if version >= 2:
            desc._read_synthdef2(stream, keep_def)
        else:
            desc._read_synthdef(stream, keep_def)
        desc.metadata = synthdef.metadata
        if keep_def:
            desc.sdef = synthdef
        return desc

    @classmethod
    def read(cls, path, keep_defs=False):
        path = pathlib.Path(path)
        ret = []
        for filename in path.parent.glob(path.name):
            with open(filename, 'rb') as file:
                ret.extend(cls._read_stream(file, keep_defs, filename))
        return ret

    @classmethod
    def _read_stream(cls, stream, keep_defs=False, path=None):  # Was readFile.
        # path is for metadata.
        stream.read(4)  # SCgf
        version = frw.read_i32(stream)
        num_defs = frw.read_i16(stream)
        ret = []
        for _ in range(num_defs):
            desc = SynthDesc()
            if version >= 2:
                desc._read_synthdef2(stream, keep_defs)
            else:
                desc._read_synthdef(stream, keep_defs)
            ret.append(desc)
            if path:
                desc.metadata = cls.md_plugin.read_file(
                    path.parent / f'{path.stem}.{type(cls.md_plugin).SUFFIX}')
            cls.populate_metadata_func(desc)
            in_memory_stream = isinstance(stream, io.BytesIO)
            if desc.sdef is not None and not in_memory_stream:
                if desc.metadata is None:
                    desc.metadata = dict()
                desc.metadata['reconstructed'] = True  # Was 'shouldNotSend'.
                desc.metadata['load_path'] = str(path)
                desc.sdef.metadata = desc.metadata
        return ret

    def _read_synthdef(self, stream, keep_def=False):  # TODO
        raise NotImplementedError(
            'read_synthdef format version 1 not implemented')

    def _read_ugen_spec(self, stream):  # TODO
        raise NotImplementedError(
            'read_ugen_spec format version 1 not implemented')

    def _read_synthdef2(self, stream, keep_def=False):
        with _libsc3.main._def_build_lock:
            try:
                self.inputs = []
                self.outputs = []
                self.control_names = []
                self.control_dict = dict()

                self.name = frw.read_pascal_str(stream)

                self.sdef = sdf.SynthDef._dummy(self.name)
                _libsc3.main._current_synthdef = self.sdef

                num_constants = frw.read_i32(stream)
                self.constants = frw.read_f32_list(stream, num_constants)

                num_controls = frw.read_i32(stream)
                self.sdef._controls = frw.read_f32_list(stream, num_controls)
                self.controls = [
                    iou.ControlName('?', i, '?', self.sdef._controls[i], None)
                    for i in range(num_controls)]

                num_control_names = frw.read_i32(stream)
                for _ in range(num_control_names):
                    control_name = frw.read_pascal_str(stream)
                    control_index = frw.read_i32(stream)
                    self.controls[control_index].name = control_name
                    self.control_names.append(control_name)
                    self.control_dict[control_name] = \
                        self.controls[control_index]

                num_ugens = frw.read_i32(stream)
                for _ in range(num_ugens):
                    self._read_ugen_spec2(stream)

                # Append all default values of each multichannel
                # control to the fist ControlName default value.
                aux_ctrl = None
                for ctrl in self.controls:
                    if ctrl.name == '?':
                        default_value = utl.as_list(aux_ctrl.default_value)
                        default_value.append(ctrl.default_value)
                        aux_ctrl.default_value = default_value
                    else:
                        aux_ctrl = ctrl

                self.sdef._control_names = [
                    x for x in self.controls if x.name is not None]
                self.has_array_args = any(
                    cn.name == '?' for cn in self.controls)

                num_variants = frw.read_i16(stream)
                self.has_variants = num_variants > 0
                # // maybe later, read in variant names and values
                # // this is harder than it might seem at first

                self.sdef._constants = dict()
                for i, k in enumerate(self.constants):
                    self.sdef._constants[k] = i

                if not keep_def:
                    # // throw away unneeded stuff
                    self.sdef = None
                    self.constats = None

                self._check_synthdesc2()
            finally:
                _libsc3.main._current_synthdef = None

    def _read_ugen_spec2(self, stream):
        ugen_class = frw.read_pascal_str(stream)
        try:
            ugen_class = ugns.installed_ugens[ugen_class]
        except NameError as e:
            raise Exception(
                f"no UGen class found for '{ugen_class}' which was "
                f"specified in synth def file: {self.name}") from e

        rate_index = frw.read_i8(stream)
        num_inputs = frw.read_i32(stream)
        num_outputs = frw.read_i32(stream)
        special_index = frw.read_i16(stream)

        # NOTE: _write_input_spec writes _synth_index and _output_index as i32.
        input_specs = frw.read_i32_list(stream, num_inputs * 2)
        output_specs = frw.read_i8_list(stream, num_outputs)  # Not used.

        ugen_inputs = []
        for i in range(0, len(input_specs), 2):
            ugen_index = input_specs[i]
            output_index = input_specs[i + 1]
            if ugen_index < 0:
                input = self.constants[output_index]
            else:
                ugen = self.sdef._children[ugen_index]
                if isinstance(ugen, ugn.MultiOutUGen):
                    input = ugen._channels[output_index]
                else:
                    input = ugen
            ugen_inputs.append(input)

        rate = self._RATE_NAME[rate_index]
        ugen = ugen_class._new_from_desc(
            rate, num_outputs, ugen_inputs, special_index)
        if isinstance(ugen, ugn.OutputProxy):
            ugen = ugen.source_ugen
        ugen._add_to_synth()

        def add_iodesc(iolst, nchan):  # lambda
            b = ugen.inputs[0]
            if type(b) is ugn.OutputProxy\
            and isinstance(b.source_ugen, iou.Control):
                control = None
                cmp_index = b._output_index + b.source_ugen._special_index
                for item in self.controls:  # detect
                    if item.index == cmp_index:
                        control = item
                        break
                if control is not None:
                    b = control.name
            iolst.append(IODesc(rate, nchan, b, ugen_class))

        if issubclass(ugen_class, iou.AbstractControl):
            # // Control.newFromDesc does not set the specialIndex, since it
            # // doesn't call Control-init. Therefore we fill it in here.
            ugen._special_index = special_index
            for i in range(num_outputs):
                self.controls[i + special_index].rate = rate
        elif issubclass(ugen_class, iou.AbstractIn):
            add_iodesc(self.inputs, len(ugen._channels))
        elif issubclass(ugen_class, iou.AbstractOut):
            add_iodesc(self.outputs, ugen._num_audio_channels())

    def _check_synthdesc2(self):
        names = set()
        nm = None

        # For reasons I don't know, synthdef controls can have duplicated
        # names, beacuse control values can be assigned by position. That is
        # not possible if the file was created from this library or sclang's
        # standard interface, unless there is a bug somewhere. For simplicity
        # and consistnecy, I'm considering the files as malformed and throwing
        # an error. This can be review later.
        for cname in self.controls:
            nm = cname.name
            if nm != '?' and nm in names:
                raise SynthDescError(
                    f"SynthDesc '{self.name}' has duplicated "
                    f"control name '{nm}'")
            else:
                names.add(nm)

        if len(names) > 255:
            raise SynthDescError(
                "a SynthDef cannot have more than 255 "
                f"control names ('{self.name}')")

        if 'gate' in names:
            self.has_gate = True

    def send(self, server, completion_msg=None):
        self.sdef.send(server, completion_msg)

    def write_metadata(self, path, md_plugin=None):
        md_plugin = md_plugin or self.md_plugin
        if self.metadata is None:
            md_plugin.delete(self.sdef, path)
        else:
            md_plugin.write(self.sdef, path)

    @classmethod
    def def_name_from_bytes(cls, data: bytearray):
        # // parse the def name out of the bytes array sent with /d_recv
        stream = io.BytesIO(data)
        stream.read(4)  # getInt32 // SCgf
        version = frw.read_i32(stream)  # Not used.
        num_defs = frw.read_i16(stream)  # Not used.
        return frw.read_pascal_str(stream)

    # def output_data(self):
    #     ugens = self.sdef._children
    #     outs = [x for x in ugens if x._writes_to_bus()]
    #     return [
    #         {'rate': x.rate, 'num_channels': x._num_audio_channels()}
    #         for x in outs]

    def __str__(self):  # Was printOn.
        string = f"SynthDesc '{self.name}':"
        for control in self.controls:
            string += f'\n  K {repr(control)}'
        for input in self.inputs:
            string += f'\n  I {repr(input)}'
        for output in self.outputs:
            string += f'\n  O {repr(output)}'
        return string


class MetaSynthDescLib(type):
    def __init__(cls, *_):
        cls.all = dict()

        def init_func(cls):
            cls.default = cls('default')  # Was global in sclang.
            sac.ServerBoot.add('all', cls.__on_server_boot)  # *** NOTE: *send calls global.send if server has booted, see below.

        clb.ClassLibrary.add(cls, init_func)


    ### System Actions ###

    def __on_server_boot(cls, server):
        # // tryToLoadReconstructedDefs = false:
        # // since this is done automatically, w/o user action,
        # // it should not try to do things that will cause warnings
        # // (or errors, if one of the servers is not local)
        if server._status_watcher.has_booted:
            cls.default.send(server, False)


class SynthDescLib(metaclass=MetaSynthDescLib):
    def __init__(self, name, servers=None):
        self.name = name
        self.all[name] = self
        self.synth_descs = dict()
        self.servers = set(servers or [srv.Server.default])

    @classmethod
    def get_lib(cls, libname):
        try:
            return cls.all[libname]
        except KeyError as e:
            raise Exception(f"library '{libname}' not found") from e

    def add(self, synth_desc):
        self.synth_descs[synth_desc.name] = synth_desc
        mdl.NotificationCenter.notify(self, 'sdesc_added', synth_desc)

    def remove_at(self, name): # BUG: es remove_at porque es un diccionario, pero es interfaz de esta clase que oculta eso, ver qué problemas puede traer.
        self.synth_descs.pop(name) #, None) # BUG: igualmente self.servers es un set y tirar KeyError con remove

    def add_server(self, server):
        self.servers.add(server)

    def remove_server(self, server):
        self.servers.remove(server)

    def at(self, name):
        return self.synth_descs[name]

    def match(self, name):
        if '.' in name:
            dot_index = name.index('.')
        else:
            return self.synth_descs[name] # BUG: tira KeyError, en sclang nil para la variable ~synthDesc puede significar otra cosa. La usa solo en PmonoStream.prInit al parecer.

        if name[:dot_index] in self.synth_descs:
            desc = self.synth_descs[name[:dot_index]]
            if desc.has_variants:
                return desc # BUG: no me cierra que no compruebe que el nombre de la variente exista, ver PmonoStream.prInit

        return self.synth_descs[name] # BUG: tira KeyError, en sclang nil para la variable ~synthDesc puede significar otra cosa. La usa solo en PmonoStream.prInit al parecer.

    def send(self, server=None, try_reconstructed=True):
        server_list = utl.as_list(server) or self.servers
        for s in server_list:
            for desc in self.synth_descs.values():
                if not desc.sdef.metadata.get('reconstructed', False):
                    desc.send(s)
                elif try_reconstructed:
                    desc.sdef._load_reconstructed(s)

    def read(self, path=None, keep_defs=True):
        path = path or plf.Platform.synthdef_dir / f'*.{sdf.SynthDef._SUFFIX}'
        path = pathlib.Path(path)
        for desc in SynthDesc.read(path, keep_defs):
            self.add(desc)
            mdl.NotificationCenter.notify(self, 'sdesc_added', desc)



"""Spec.sc"""

from collections import namedtuple

from ..base import builtins as bi


__all__ = ['spec']


class Warp():
    _specifier = None
    __slots__ = ('_spec',)

    def __init__(self, spec):
        self._spec = spec

    @property
    def specifier(self):
        return type(self)._specifier

    def map(self, value):
        return value

    def unmap(self, value):
        return value

    def __eq__(self, other):
        if type(self) == type(other):
            return True
        else:
            return False

    def __hash__(self):
        return hash(type(self))


class LinearWarp(Warp):
    _specifier = 'lin'

    def map(self, value):
        return value * self._spec.range + self._spec.minval

    def unmap(self, value):
        return (value - self._spec.minval) / self._spec.range


class ExponentialWarp(Warp):
    # // Both minval and maxval must be non zero and have the same sign.

    _specifier = 'exp'

    def map(self, value):
        return self._spec.ratio ** value * self._spec.minval

    def unmap(self, value):
        return bi.log(value / self._spec.minval) / bi.log(self._spec.ratio)


class CurveWarp(Warp):
    __slots__ = ('_spec', '_curve', '_a', '_b', '_grow')

    def __new__(cls, spec, curve):
        # // Prevent math blow up.
        if abs(curve) < 0.001:
            return LinearWarp(spec)
        return super().__new__(cls)

    def __init__(self, spec, curve):
        self._spec = spec
        if curve > -0.001 and curve < 0.001:
            self._curve = 0.001
        else:
            self._curve = curve
        self._grow = bi.exp(self._curve)
        self._a = spec.range / (1.0 - self._grow)
        self._b = spec.minval + self._a

    @property
    def specifier(self):
        return self._curve

    @property
    def curve(self):
        return self._curve

    def map(self, value):
        return self._b - self._a * pow(self._grow, value)

    def unmap(self, value):
        return bi.log((self._b - value) / self._a) / self._curve

    def __eq__(self, other):
        if type(self) == type(other):
            return self.curve == other.curve
        else:
            return False

    def __hash__(self):
        return hash(type(self))


class CosineWarp(LinearWarp):
    _specifier = 'cos'

    def map(self, value):
        return super().map(0.5 - bi.cos(bi.pi * value) * 0.5)

    def unmap(self, value):
        return bi.acos(1.0 - super().unmap(value) * 2.0) / bi.pi


class SineWarp(LinearWarp):
    _specifier = 'sin'

    def map(self, value):
        return super().map(bi.sin(bi.pi2 * value))

    def unmap(self, value):
        return bi.asin(super().unmap(value)) / bi.pi2


class AmpWarp(Warp):
    _specifier = 'amp'

    def map(self, value):
        if self._spec.range >= 0:
            return value ** 2 * self._spec.range + self._spec.minval
        else:
            # // The formula can be reduced to (2*v) - v.squared
			# // but the 2 subtractions would be faster.
            return (1 - (1 - value) ** 2) * self._spec.range + self._spec.minval

    def unmap(self, value):
        if self._spec.range >= 0:
            return bi.sqrt((value - self._spec.minval) / self._spec.range)
        else:
            return (
                1 - bi.sqrt(1 - (value - self._spec.minval) / self._spec.range))


class DbWarp(Warp):
    _specifier = 'db'

    def map(self, value):
        dbamp = bi.dbamp
        range = dbamp(self._spec.maxval) - dbamp(self._spec.minval)
        if self._spec.range >= 0:
            return bi.ampdb(value ** 2 * range + dbamp(self._spec.minval))
        else:
            return bi.ampdb(
                (1 - (1 - value) ** 2) * range + dbamp(self._spec.minval))

    def unmap(self, value):
        dbamp = bi.dbamp
        if self._spec.range >= 0:
            return bi.sqrt(
                (dbamp(value) - dbamp(self._spec.minval)) /
                (dbamp(self._spec.maxval) - dbamp(self._spec.minval)))
        else:
            return 1 - bi.sqrt(
                1 - (dbamp(value) - dbamp(self._spec.minval)) /
                (dbamp(self._spec.maxval) - dbamp(self._spec.minval)))


# Warp *initClass

_WARPS = {
    LinearWarp._specifier: LinearWarp,
    ExponentialWarp._specifier: ExponentialWarp,
    SineWarp._specifier: SineWarp,
    CosineWarp._specifier: CosineWarp,
    AmpWarp._specifier: AmpWarp,
    DbWarp._specifier: DbWarp
    # // CurveWarp is specified by a number.
}

def _as_warp(warp, spec):
    if isinstance(warp, str):
        try:
            return _WARPS[warp](spec)
        except KeyError:
            raise ValueError(f"invalid warp '{warp}'")
    elif isinstance(warp, (int, float)):
        return CurveWarp(spec, warp)
    elif isinstance(warp, Warp):
        return warp
    else:
        raise TypeError(f"can't create a warp from {type(warp).__name__}")


class ControlSpec():
    def __init__(self, minval=0.0, maxval=1.0, warp='lin',
                 step=None, default=None, units=None):  #, grid=None):
        self._minval = minval
        self._maxval = maxval
        self._step = 0.0 if step is None else step
        self._default = minval if default is None else default
        self._units = '' if units is None else units
        self._range = maxval - minval
        self._ratio = maxval / minval if minval != 0 else float('inf')
        self._cliplo = min(minval, maxval)
        self._cliphi = max(minval, maxval)
        self._warp = _as_warp(warp, self)
        # self.grid = grid
        self._hash = hash((type(self), *vars(self).values()))

    # newFrom
    # copy

    @property
    def minval(self):
        return self._minval

    @property
    def maxval(self):
        return self._maxval

    @property
    def warp(self):
        return self._warp

    @property
    def step(self):
        return self._step

    @property
    def default(self):
        return self._default

    @property
    def units(self):
        return self._units

    @property
    def range(self):
        return self._range

    @property
    def ratio(self):
        return self._ratio

    def map(self, value):
        # // Maps a value from [0..1] to spec range.
        return self._warp.map(bi.round(bi.clip(value, 0.0, 1.0), self._step))

    def unmap(self, value):
        # // Maps a value from spec range to [0..1].
        return self._warp.unmap(
            bi.clip(bi.round(value, self._step), self._cliplo, self._cliphi))

    # GUI specific methods:
    # constrain
    # guessNumberStep
    # grid
    # looseRange
    # normalize
    # roundRange
    # gridValues
    # zoom
    # shift
    # setFrom

    def __eq__(self, other):
        if type(self) == type(other):
            return vars(self) == vars(other)
        else:
            return False

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return (
            f"{type(self).__name__}({self._minval}, {self._maxval}, "
            f"{repr(self._warp.specifier)}, {self._step}, {self._default}, "
            f"'{self._units}')")


# ControlSpec *initClass

_SPECS = {
    'unipolar': ControlSpec(0, 1),
    'bipolar': ControlSpec(-1, 1, default=0),

    'freq': ControlSpec(20, 20000, 'exp', 0, 440, units='Hz'),
    'lofreq': ControlSpec(0.1, 100, 'exp', 0, 6, units='Hz'),
    'midfreq': ControlSpec(25, 4200, 'exp', 0, 440, units='Hz'),
    'widefreq': ControlSpec(0.1, 20000, 'exp', 0, 440, units='Hz'),
    'phase': ControlSpec(0, bi.twopi),
    'rq': ControlSpec(0.001, 2, 'exp', 0, 0.707),

    'midi': ControlSpec(0, 127, default=64),
    'midinote': ControlSpec(0, 127, default=60),
    'midivelocity': ControlSpec(1, 127, default=64),

    'db': ControlSpec(bi.ampdb(0), bi.ampdb(1), 'db', units='dB'),
    'amp': ControlSpec(0, 1, 'amp', 0, 0),
    'boostcut': ControlSpec(-20, 20, units='dB', default=0),

    'pan': ControlSpec(-1, 1, default=0),
    'detune': ControlSpec(-20, 20, default=0, units='Hz'),
    'rate': ControlSpec(0.125, 8, 'exp', 0, 1),
    'beats': ControlSpec(0, 20, units='Hz'),

    'delay': ControlSpec(0.0001, 1, 'exp', 0, 0.3, units='secs')
}

def spec(spc):
    if isinstance(spc, ControlSpec):
        return spc
    elif isinstance(spc, str):
        try:
            return _SPECS[spc]
        except KeyError:
            raise ValueError(f"invalid spec '{spc}'") from None
    elif isinstance(spc, (int, float)):
        return ControlSpec(warp=spc)
    elif isinstance(spc, (list, tuple)):
        return ControlSpec(*spc)
    else:
        raise TypeError(f"can't create a spec from {type(spc).__name__}")

# add_spec/remove_spec might not be a good idea.


### Metadata support ###

def key_encoder(dct):
    # Receives a dict of ControlSpecs, the 'specs' key of SynthDef
    # metadata and return JSON serializable Python objects.
    return {
        k: [v._minval, v._maxval, v._warp.specifier,
            v._step, v._default, v._units]
        for k, v in dct.items()}

def key_decoder(dct):
    # Receives a dict of lists (from a JSON file) for the 'specs'
    # key of SynthDef and return a dict of ControlSpecs.
    return {k: spec(v) for k, v in dct.items()}

SpecCodec = namedtuple('SpecCodec', ['key', 'encoder', 'decoder'])
spec_codec = SpecCodec('specs', key_encoder, key_decoder)




from . import builtins as bi
from . clock import *
from . functions import *
from . main import main
from . model import *
from . netaddr import *
from . operand import *
from . platform import *
from . play import *
from . responsedefs import *
from . stream import *
from . systemactions import *
from . import utils as utl



"""MacUGens.sc"""

from .. import ugen as ugn


class MouseX(ugn.UGen):
    _default_rate = 'control'
    _WARPS = {
        'lin': 0, 'linear': 0, 0: 0,
        'exp': 1, 'exponential': 1, 1: 1}

    @classmethod
    def kr(cls, minval=0, maxval=1, warp=0, lag=0.2):
        try:
            warp = cls._WARPS[warp]
        except KeyError:
            raise ValueError(f'invalid warp {repr(warp)}') from None
        return cls._multi_new('control', minval, maxval, warp, lag)

    @classmethod
    def signal_range(cls):  # override
        return 'unipolar'


class MouseY(MouseX):
    pass


class MouseButton(ugn.UGen):
    _default_rate = 'control'

    @classmethod
    def kr(cls, minval=0, maxval=1, lag=0.2):
        return cls._multi_new('control', minval, maxval, lag)

    @classmethod
    def signal_range(cls):  # override
        return 'unipolar'


class KeyState(ugn.UGen):
    _default_rate = 'control'

    @classmethod
    def kr(cls, keycode=0, minval=0, maxval=1, lag=0.2):
        return cls._multi_new('control', keycode, minval, maxval, lag)

    @classmethod
    def signal_range(cls):  # override
        return 'unipolar'



"""Bus.sc"""

import logging

from . import _graphparam as gpp
from . import server as srv
from ..base import utils as utl
from ..base import responsedefs as rdf


__all__ = ['AudioBus', 'ControlBus']


_logger = logging.getLogger(__name__)


class BusException(Exception):
    pass


class BusAlreadyFreed(BusException):
    def __init__(self, method=None):
        if method is not None:
            self.args = (f"'{method}' method called",)


class Bus(gpp.UGenParameter, gpp.NodeParameter):
    '''Client side representation of server buses.

    Bus objects are used to keep track and manage the buses being
    used in the server, they can be passed as arguments instead
    of using the bus number directly.

    From ControlBus objects bus values can be obtained or assigned
    via OSC commands.

    Parameters
    ----------
    num_channels : int
        Number of channels, default is 1.
    server : Server
        Target server.
    index : int
        Sets the bus index. If ``index`` is None bus number
        is automatically allocated (client side) by the bus
        allocator class.
    '''

    # @classmethod
    # def audio(cls, channels=1, server=None, index=None):
    #     return AudioBus(channels, server, index)
    #
    # @classmethod
    # def control(cls, channels=1, server=None, index=None):
    #     return ControlBus(channels, server, index)

    @property
    def rate(self):
        return self._rate

    @property
    def index(self):
        return self._index

    @property
    def channels(self):
        return self._channels

    @property
    def server(self):
        return self._server

    @classmethod
    def new_from(cls, bus, offset, channels=1):
        if offset > bus._channels\
        or channels + offset > bus._channels:
            raise BusException(
                'new_from tried to reach outside '
                f'the channel range of {bus}')
        return cls(channels, bus.server, bus.index + offset)

    def free(self, clear=False):
        if self._index is None:
            _logger.warning('bus has already been freed')
            return
        if self._rate == 'audio':
            self._server._audio_bus_allocator.free(self._index)
        else:
            self._server._control_bus_allocator.free(self._index)
            if clear:
                self.fill(0, self._channels)
        self._index = None
        self._channels = None
        self._map_symbol = None

    # # Allow reallocation. Don't allow, use instances.
    #
    # def alloc(self):
    #     if self._rate == 'audio':
    #         self._index = self._server._audio_bus_allocator.alloc(
    #             self._channels)
    #     else:
    #         self._index = self._server._control_bus_allocator.alloc(
    #             self._channels)
    #     self._map_symbol = None
    #
    # def realloc(self):
    #     if self._index is None:
    #         raise BusAlreadyFreed('realloc')
    #     rate = self._rate
    #     num_channels = self._channels
    #     self.free()
    #     self._rate = rate
    #     self._channels = num_channels
    #     self.alloc()

    # setAll is fill(value, self._channels)
    # value_ is fill(value, self._channels)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        if self._index == other._index\
        and self._channels == other._channels\
        and self._rate == other._rate\
        and self._server == other._server:
            return True
        else:
            return False

    def __hash__(self):
        return hash(
            (type(self), self._index, self._channels,
            self._rate, self._server))

    def __repr__(self):
        return (
            f'{type(self).__name__}({self._channels}, '
            f'{self._index}, {self._server.name})')

    def as_map(self):
        if self._map_symbol is None:
            if self._index is None:
                raise BusException('bus not allocated')
            self._map_symbol = 'a' if self._rate == 'audio' else 'c'
            self._map_symbol += str(self._index)
        return self._map_symbol

    def sub_bus(self, offset, num_channels=1):
        return type(self).new_from(self, offset, num_channels)

    # ar
    # kr
    # play


    ### UGen graph parameter interface ###

    def _as_ugen_input(self, *_):
        return self._index

    def _as_ugen_rate(self):
        return self._rate


    ### Node parameter interface ###

    def _as_control_input(self):
        return self._index


class AudioBus(Bus):
    def __init__(self, channels=1, server=None, index=None):
        super(gpp.UGenParameter, self).__init__(self)
        self._channels = channels
        self._server = server or srv.Server.default
        self._rate = 'audio'
        if index is None:
            self._index = self._server._audio_bus_allocator.alloc(
                self._channels)
            if self._index is None:
                raise BusException(
                    'failed to get AudioBus bus allocated, channels '
                    f"= {self._channels}, server = {self._server.name}")
        else:
            self._index = index
        self._map_symbol = None

    # @property
    # def is_audio_io(self):
    #     return self._index < self._server.options.first_private_bus()


    ### Shared memory interface ###

    # def get_synchronous(self):
    #     ...

    # def getn_synchronous(self, count):
    #     ...

    # def set_synchronous(self, *values):
    #     ...

    # def setn_synchronous(self, values):
    #     ...


class ControlBus(Bus):
    def __init__(self, channels=1, server=None, index=None):
        super(gpp.UGenParameter, self).__init__(self)
        self._channels = channels
        self._server = server or srv.Server.default
        self._rate = 'control'
        if index is None:
            self._index = self._server._control_bus_allocator.alloc(
                self._channels)
            if self._index is None:
                raise BusException(
                    'failed to get ControlBus bus allocated, channels '
                    f'= {self._channels}, server = {self._server.name}')
        else:
            self._index = index
        self._map_symbol = None

    def set(self, *values):  # // shouldn't be larger than self._channels
        if self._index is None:
            raise BusAlreadyFreed('set')
        msg = ['/c_set']
        values = [[self._index + i, val] for i, val in enumerate(values)]
        msg.extend(utl.flat(values))
        self._server.send_bundle(0, msg)

    def set_msg(self, *values):
        if self._index is None:
            raise BusAlreadyFreed('set_msg')
        msg = ['/c_set']
        values = [[self._index + i, val] for i, val in enumerate(values)]
        msg.extend(utl.flat(values))
        return msg

    def setn(self, values):
        if self._index is None:
            raise BusAlreadyFreed('setn')
        msg = ['/c_setn', self._index, len(values)]
        msg.extend(values)
        self._server.send_bundle(0, msg)

    def setn_msg(self, values):
        if self._index is None:
            raise BusAlreadyFreed('setn_msg')
        msg = ['/c_setn', self._index, len(values)]
        msg.extend(values)
        return msg

    def set_at(self, offset, *values):
        if self._index is None:
            raise BusAlreadyFreed('set_at')
        msg = ['/c_set']
        values = [
            [self._index + offset + i, val] for i, val in enumerate(values)]
        msg.extend(utl.flat(values))
        self._server.send_bundle(0, msg)

    def setn_at(self, offset, values):
        if self._index is None:
            raise BusAlreadyFreed('setn_at')
        # // could throw an error if values.size > numChannels
        msg = ['/c_setn', self._index + offset, len(values)]
        msg.extend(values)
        self._server.send_bundle(0, msg)

    def set_pairs(self, *pairs):
        if self._index is None:
            raise BusAlreadyFreed('set_pairs')
        msg = ['/c_set']
        pairs = [[self._index + pair[0], pair[1]]\
                 for pair in utl.gen_cclumps(pairs, 2)]
        msg.extend(utl.flat(pairs))
        self._server.send_bundle(0, msg)

    def get(self, action=None):
        if self._index is None:
            raise BusAlreadyFreed('get')

        if self._channels == 1:
            if action is None:
                def default_action_func(val):
                    print(
                        f'{type(self).__name__} index: '
                        f'{self._index} value: {val}')
                action = default_action_func

            def get_func(msg, *_):
                # // The response is of the form [/c_set, index, value].
                # // We want "value," which is at index 2.
                action(msg[2])

            rdf.OscFunc(
                get_func, '/c_set', self._server.addr,
                arg_template=[self._index]).one_shot()
            self._server.send_msg('/c_get', self._index)
        else:
            self.getn(self._channels, action)

    def get_msg(self):
        if self._index is None:
            raise BusAlreadyFreed('get_msg')
        return ['/c_get', self._index]

    def getn(self, count=None, action=None):
        if self._index is None:
            raise BusAlreadyFreed('getn')

        if action is None:
            def default_action_func(vals):
                print(
                    f'{type(self).__name__} index: '
                    f'{self._index} values: {vals}')
            action = default_action_func

        def getn_func(msg, *_):
            # // The response is of the form [/c_set, index, count, ...values].
            # // We want the values, which are at indexes 3 and above.
            action(msg[3:])

        rdf.OscFunc(
            getn_func, '/c_setn', self._server.addr,
            arg_template=[self._index]).one_shot()
        if count is None:
            count = self._channels
        self._server.send_msg('/c_getn', self._index, count)

    def getn_msg(self, count=None):
        if self._index is None:
            raise BusAlreadyFreed('getn_msg')
        if count is None:
            count = self._channels
        return ['/c_getn', self._index, count]

    def fill(self, value, num_channels):
        if self._index is None:
            raise BusAlreadyFreed('fill')
        # // Could throw an error if numChans > numChannels.
        self._server.send_bundle(
            None, ['/c_fill', self._index, num_channels, value])

    def fill_msg(self, value):
        if self._index is None:
            raise BusAlreadyFreed('fill_msg')
        return ['/c_fill', self._index, self._channels, value]



"""Engine.sc"""

import math
import random
import logging

from ..base import builtins as bi


_logger = logging.getLogger(__name__)


class NodeIDAllocator():
    def __init__(self, user=0, init_temp=1000):
        if user > 31:
            raise Exception("NodeIDAllocator user id > 31")
        self.user = user
        self._init_temp = init_temp
        self.num_ids = (2 ** 32 // 2 - 1) // 64  # +int32, 64 logins safe range.
        self.reset()

    def id_offset(self):
        return self.num_ids * self.user

    def reset(self):
        self._mask = self.user << 26
        self._temp = self._init_temp
        self._perm = 2
        self._perm_freed = set()

    def alloc(self):
        x = self._temp
        self._temp = bi.wrap(x + 1, self._init_temp, 0x03FFFFFF)
        return x | self._mask

    def alloc_perm(self):
        if len(self._perm_freed) > 0:
            x = min(self._perm_freed)
            self._perm_free.remove(x)
        else:
            x = self._perm
            self._perm = min(x + 1, self._init_temp - 1)
        return x | self.mask

    def free_perm(self, id):
        # // should not add a temp node id to the freed-permanent collection
        id = id & 0x03FFFFFF
        if id < self._init_temp:
            self._perm_freed.add(id)


class PowerOfTwoBlock():
    def __init__(self, addr, size):
        self.addr = addr
        self.size = size
        self.next = None


class PowerOfTwoAllocator():
    # // THIS IS THE RECOMMENDED ALLOCATOR FOR BUSES AND BUFFERS
    def __init__(self, size, pos=0):
        self._size = size
        self._array = [None] * size
        self._free_lists = [None] * 32
        self._pos = pos

    def alloc(self, n):
        # TODO: NEXTPOWEROFTWO está en: /include/common/clz.h
        # nextPowerOf en sclang: pow(2, math.ceil(math.log(x) / math.log(2))) -> int
        n = pow(2, math.ceil(math.log(n) / math.log(2))) # nextPowerOfTwo
        size_class = math.ceil(math.log2(n)) # log2Ceil primitive
        node = self._free_lists[size_class]
        if node is not None:
            self._free_lists[size_class] = node.next
            return node.addr
        if self._pos + n <= self._size:
            self._array[self._pos] = PowerOfTwoBlock(self._pos, n)
            addr = self._pos
            self._pos += n
            return addr
        return None

    def free(self, addr):
        # BUG: typo, declara y no usa la variable next (que no es la propiedad del bloque)
        node = self._array[addr]
        if node is not None:
            size_class = math.ceil(math.log2(node.size)) # log2Ceil primitive
            node.next = self._free_lists[size_class]
            self._free_lists[size_class] = node
            self._array[addr] = None

    def blocks(self):
        return [x for x in self._array if x is not None]


class LRUNumberAllocator():
    # // implements a least recently used ID allocator.
    def __init__(self, lo, hi):
        self._lo = lo
        self._hi = hi
        self._init()

    def _init(self):
        self._size = self._hi - self._lo + 1
        self._array = [None] * self._size
        for i, j in enumerate(range(self._lo, self._hi)):
            self._array[i] = j
        self._head = self._size - 1
        self._tail = 0

    def alloc(self):
        if self._head == self._tail:
            return None # // empty
        id = self._array[self._tail]
        self._tail = (self._tail + 1) % self._size
        return id

    def free(self, id):
        next_index = (self._head + 1) % self._size
        if next_index == self._tail:
            return None # // full
        self._array[self._head] = id
        self._head = next_index


class StackNumberAllocator():
    def __init__(self, lo, hi):
        self._lo = lo
        self._hi = hi
        self._init()

    def _init(self):
        self._free_list = []
        self._next = self._lo - 1

    def alloc(self):
        if len(self._free_list) > 0:
            return self._free_list.pop()
        if self._next < self._hi:
            self._next += 1
            return self._next
        return None

    def free(self, index):
        self._free_list.append(index)


class RingNumberAllocator():
    def __init__(self, lo, hi):
        self._lo = lo
        self._hi = hi
        self._init()

    def _init(self):
        self._next = self._hi

    def alloc(self):
        self._next = bi.wrap(self._next + 1, self._lo, self._hi)
        return self._next


# // by hjh: for better handling of dynamic allocation

class ContiguousBlock():
    def __init__(self, start, size):
        self.start = start
        self.size = size
        self.used = False # // assume free; owner must say otherwise

    @property
    def address(self):
        return self.start

    def adjoins(self, block):
        return (self.start < block.start and self.start + self.size >= block.start)\
        or (self.start > block.start and block.start + block.size >= self.start)

    def join(self, block):
        if self.adjoins(block):
            start = min(self.start, block.start)
            size = max(self.start + self.size, block.start + block.size) - start
            return type(self)(start, size)
        else:
            return None

    def split(self, span):
        if span < self.size:
            return [
                type(self)(self.start, span),
                type(self)(self.start + span, self.size - span)
            ]
        elif span == self.size:
            return [self, None]
        else:
            return [None, None]

    # storeArgs { ^[start, size, used] }
    # TODO (en general) printOn usa sotreOn que usa storeParamsOn que
    # usa storeArgs. Acá lo puse todo en __repr__ manualmente.
    # BUG: el problema es que se usa inconsistentemente en la librería
    # de clases. En Control se usa como __str__ y acá se usa como __repr__.
    # BUG: y también acá está mal usado porque la cadena de printOn
    # llama a obj.simplifyStoreArgs que no filtra los argumentos, o no
    # se usa, correctamente porque devuelve una llamada al constructor
    # con un argumento de más ('used').
    # def __repr__(self):
    #     string = type(self).__name__
    #     string += '({}, {})'.format(self.start, self.size)
    #     return string


# TODO: comentario duplicado en sclang...
class ContiguousBlockAllocator():
    # // pos is offset for reserved numbers,
    # // addrOffset is offset for clientID * size
    def __init__(self, size, pos=0, addr_offset=0):
        self.size = size
        shifted_pos = pos + addr_offset
        self._array = [None] * size
        self._array[pos] = ContiguousBlock(shifted_pos, size - pos)
        self._freed = dict()
        self.pos = shifted_pos
        self.top = shifted_pos
        self.addr_offset = addr_offset

    def alloc(self, n=1):
        block = self._find_available(n)
        if block is not None:
            return self._reserve(block.start, n, block).start
        else:
            return None

    def reserve(self, addr, size=1, warn=True):
        if self._array[addr] is None:
            block = self._find_next(addr)
        else:
            block = self._array[addr]
        if block is not None and block.used and addr + size > block.start:
            if warn:
                _logger.warning(f'The block at ({addr}, {size}) is '
                                'already in use and cannot be reserved.')
        elif block.start == addr:
            return self._reserve(addr, size, block)

        block = self._find_previous(addr)
        if block is not None and block.used and block.start + block.size > addr:
            if warn:
                _logger.warning(f'The block at ({addr}, {size}) is '
                                'already in use and cannot be reserved')
        else:
            return self._reserve(addr, size, None, block)

        return None

    def free(self, addr):
        # // this 'if' prevents an error if a Buffer object is freed twice
        if addr is None:
            return
        block = self._array[addr - self.addr_offset]
        if block is not None and block.used:
            block.used = False
            self._add_to_freed(block)
            prev = self._find_previous(addr)
            if prev is not None and not prev.used:
                tmp = prev.join(block)
                if tmp is not None:
                    # // if block is the last one, reduce the top
                    if block.start == self.top: self.top = tmp.start
                    self._array[tmp.start - self.addr_offset] = tmp
                    self._array[block.start - self.addr_offset] = None
                    self._remove_from_freed(prev)
                    self._remove_from_freed(block)
                    if self.top > tmp.start: self._add_to_freed(tmp)
                    block = tmp
            next = self._find_next(block.start)
            if next is not None and not next.used:
                tmp = next.join(block)
                if tmp is not None:
                    # // if next is the last one, reduce the top
                    if next.start == self.top: self.top = tmp.start
                    self._array[tmp.start - self.addr_offset] = tmp
                    self._array[next.start - self.addr_offset] = None
                    self._remove_from_freed(next)
                    self._remove_from_freed(block)
                    if self.top > tmp.start: self._add_to_freed(tmp)

    def blocks(self):
        return [x for x in self._array if x is not None and x.used]

    def _find_available(self, n):
        if len(self._freed) > 0:
            return random.choice(list(self._freed[n])) # BUG: ver las funciones de librería como esta
        for size, set_ in self._freed.items():
            if size >= n and len(set_) > 0:
                return random.choice(list(set_))
        if self.top + n - self.addr_offset > self.size\
        or self._array[self.top - self.addr_offset].used:
            return None
        return self._array[self.top - self.addr_offset]

    def _add_to_freed(self, block):
        if self._freed.get(block.size) is None:
            self._freed[block.size] = set()
        self._freed[block.size].add(block)

    def _remove_from_freed(self, block):
        if self._freed.get(block.size) is not None:
            if block in self._freed[block.size]:
                self._freed[block.size].remove(block)
            if len(self._freed[block.size]) == 0:
                del self._freed[block.size]

    def _find_previous(self, addr):
        for i in reversed(range(self.pos, addr)):
            if self._array[i - self.addr_offset] is not None:
                return self._array[i - self.addr_offset]
        return None

    def _find_next(self, addr):
        tmp = self._array[addr - self.addr_offset]
        if tmp is not None:
            return self._array[tmp.start + tmp.size - self.addr_offset]
        else:
            assert(addr < self.top) # BUG: creo tiene que ser menor que top, y la función es find 'next', pero no entiendo por qué usa sclang 'for' que hace reverse automáticamente y top es inclusive
            for i in range(addr + 1, self.top + 1):
                if self._array[i - self.addr_offset] is not None:
                    return self._array[i - self.addr_offset]
        return None

    def _reserve(self, addr, size, avail_block=None, prev_block=None):
        if avail_block is None and prev_block is None:
            prev_block = self._find_previous(addr)
        if avail_block is None:
            avail_block = prev_block
        if avail_block.start < addr:
            _, avail_block = self._split(avail_block, addr - avail_block.start, False)
        return self._split(avail_block, size, True)[0]

    def _split(self, avail_block, n, used=True):
        new, leftover = avail_block.split(n)
        new.used = used
        self._remove_from_freed(avail_block)
        if not used:
            self._add_to_freed(new)
        self._array[new.start - self.addr_offset] = new
        if leftover is not None:
            self._array[leftover.start - self.addr_offset] = leftover
            self.top = max(self.top, leftover.start)
            if self.top > leftover.start:
                self._add_to_freed(leftover)
        return [new, leftover]

    # debug { |text|
    #     Post << text << ":\n\nArray:\n";
    #     array.do({ |item, i|
    #         item.notNil.if({ Post << i << ": " << item << "\n"; });
    #     });
    #     Post << "\nFree sets:\n";
    #     freed.keysValuesDo({ |size, set|
    #         Post << size << ": " <<< set << "\n";
    #     });
    # }



"""Stream.sc part 2"""

import logging

from ..base import main as _libsc3
from ..base import stream as stm
from ..base import systemactions as sac
from ..base import clock as clk
from . import event as evt


__all__ = []


_logger = logging.getLogger(__name__)


### Higher level abstractions ###


# class OneShotStream(Stream): ... # TODO: ver para qué sirve, la única referencia está en Object:iter, no está documentada.
# class EmbedOnce(Stream): ... # TODO, ver, solo se usa en JITLib, no está documentada.
# class StreamClutch(Stream): ... # TODO: no se usa en la librería de clases, actúa como un filtro, sí está documentada.
# class CleanupStream(Stream): ... # TODO: no se usa en la librería de clases, creo, ver bien, no tiene documentación.


### EventStreamCleanup.sc ###


class CleanupEntry():
    def __init__(self):
        thread_player = _libsc3.main.current_tt.thread_player
        if isinstance(thread_player, EventStreamPlayer):
            self._cleanup = thread_player.cleanup
            self._cleanup.add(self)
        else:
            self._cleanup = EventStreamCleanup()
        self._events = set()
        self._functions = dict()

    def add_event(self, event):
        self._events.add(event)

    def remove_event(self, event):
        self._events.discard(event)

    def add_function(self, fn, *args):
        self._functions[fn] = args

    def remove_function(self, fn):
        self._functions.pop(fn, None)

    def run(self):
        for event in self._events:
            event.play()
        for fn, args in self._functions.copy().items():
            fn(*args)
        self.clear()
        if self._cleanup:
            self._cleanup.remove(self)
            self._cleanup = None

    def clear(self):
        self._events = set()
        self._functions = dict()


class EventStreamCleanup():
    def __init__(self):
        self._entries = set()
        sac.CmdPeriod.add(self.__on_cmd_period)

    def add(self, entry):
        if isinstance(entry, CleanupEntry):
            self._entries.add(entry)
        else:
            raise TypeError('entry is not a CleanupEntry')

    def remove(self, entry):
        self._entries.discard(entry)

    def clear(self):
        self._entries = set()

    def run(self):
        for entry in self._entries.copy():
            entry.run()
        self.clear()
        sac.CmdPeriod.remove(self.__on_cmd_period)


    ### System Actions ###

    def __on_cmd_period(self):  # Instead of removed_from_scheduler.
        self.run()



class EventStreamPlayer(stm.Routine):
    def __init__(self, stream, event=None):
        super().__init__(self._stream_player_func())
        self._stream = stream
        self._event = event or evt.event()
        self._is_muted = False
        self._cleanup = EventStreamCleanup()

    @property
    def cleanup(self):
        return self._cleanup

    def mute(self):
        self._is_muted = True

    def unmute(self):
        self._is_muted = False

    def reset(self):
        with self._state_cond:
            super().reset()
            self._stream.reset()
            self._cleanup.run()

    def resume(self, clock=None, quant=None):
        with self._state_cond:
            if self.state == self.State.Paused:
                self.state = self.State.Suspended
                clock = clock or self._clock or clk.SystemClock
                self._event, quant = self._synch_with_quant(self._event, quant)
                clock.play(self, quant)

    def stop(self):
        with self._state_cond:
            super().stop()
            self._cleanup.run()

    def _stream_player_func(self):
        def esp_func():
            try:
                while True:
                    outevent = self._stream.next(self._event.copy())
                    yield self._play_and_delta(outevent)
            except stm.StopStream:
                self._cleanup.run()

        return esp_func

    def _play_and_delta(self, outevent):  # Was Event.playAndDelta.
        if not (self._is_muted or evt.is_rest(outevent)):
            outevent.play()
        return outevent('delta')

    def play(self, clock=None, quant=None, reset=False):
        if reset:
            self.reset()
        with self._state_cond:
            if self.state == self.State.Init\
            or self.state == self.state.Paused:
                self.state = self.State.Suspended
                clock = clock or clk.SystemClock
                self._event, quant = self._synch_with_quant(self._event, quant)
                clock.play(self, quant)
        # Pattern.play return the stream, maybe for API usage constency
        # Stream.play should return self, but I'm not sure.
        # return self

    @staticmethod
    def _synch_with_quant(event, quant):  # Was Event.synchWithQuant.
        quant = clk.Quant.as_quant(quant)  # Also needs a copy or quant is disposable?
        if quant.timing_offset is None:
            quant.timing_offset = event('timing_offset')
        else:
            event = event.copy()
            event['timing_offset'] = quant.timing_offset
        return event, quant


### Pattern Streams ###


class PatternValueStream(stm.Stream):
    def __init__(self, pattern):
        self.pattern = pattern
        self._stream = None


    ### Stream protocol ###

    def next(self, inval=None):
        try:
            if self._stream is None:
                self._stream = self.pattern.__embed__(inval)
                return next(self._stream)
            else:
                return self._stream.send(inval)
        except StopIteration:
            raise stm.StopStream from None

    def reset(self):
        self._stream = None


class PatternEventStream(PatternValueStream):
    ### Stream protocol ###

    def next(self, inevent=None):
        try:
            inevent = evt.event() if inevent is None else inevent
            if self._stream is None:
                self._stream = self.pattern.__embed__(inevent)
                return next(self._stream)
            else:
                return self._stream.send(inevent)
        except StopIteration:
            raise stm.StopStream from None

    def reset(self):
        self.clear()
        self._stream = None

    def clear(self):
        if self._stream:
            try:
                self._stream.throw(stm.StopStream)
            except StopIteration:
                pass



"""
Builtin Pattern classes package.
"""

from ...base import _hooks as hks


lst = list(hks.import_all_module('sc3.seq.pattern', bind=__name__).keys())
lst.extend(hks.import_all_package(__path__, __name__, bind=__name__).keys())

__all__ = lst



"""SC_InlineUnaryOp.h, SC_InlineBinaryOp.h"""

# Note: Functions in this module are quite slow. They are Python
# implementations of sc server opcodes to obtain special index symbols and
# conform AbstractObject interface. A few are also wrappers to match
# sclang's behaviour too (e.g. mod and random functions). They will need to be
# optimized along with events, patterns and streams. However, when running the
# library on pypy all unoptimized code at least matches sclang.

import math
import builtins

from . import classlibrary as clb


clb.ClassLibrary.late_imports(__name__, ('sc3.base.main', '_libsc3'))


# Module Constants
# /include/plugin_interface/SC_Constants.h
pi = math.acos(-1.)
pi2 = pi * .5
pi32 = pi * 1.5
twopi = pi * 2.
rtwopi = 1. / twopi
log001 = math.log(0.001)
log01 = math.log(0.01)
log1 = math.log(0.1)
rlog2 = 1. / math.log(2.)
sqrt2 = math.sqrt(2.)
rsqrt2 = 1. / sqrt2


# There is a thing, some operators are unary or binary with argument, some are
# math functions with arguments, the problem is that when declared as binops
# they will use rcompose but unops and narops will not. That creates a possible
# behaviour inconsistency problem.
class scbuiltin():
    # def __new__(cls, func):
    #     def scbuiltin_(*args):
    #         return func(*args)
    #
    #     scbuiltin_.__name__ = func.__name__  # used to obtain special_index.
    #     scbuiltin_.__qualname__ += func.__name__
    #     return scbuiltin_

    @staticmethod
    def unop(func):
        def scbuiltin_(x):
            try:
                return x._compose_unop(func)
            except AttributeError:
                try:
                    return func(x)
                except TypeError:
                    pass
            raise TypeError(f"scbuiltin '{func.__name__}' function is not "
                            f"supported for type '{type(x).__name__}'")

        scbuiltin_.__name__ = func.__name__  # used to obtain special_index.
        scbuiltin_.__qualname__ += func.__name__
        return scbuiltin_

    @staticmethod
    def binop(func):
        def scbuiltin_(a, *b):
            try:
                return a._compose_binop(func, *b)
            except AttributeError:
                try:
                    return b._rcompose_binop(func, a)
                except AttributeError:
                    try:
                        return func(a, *b)
                    except TypeError:
                        pass
            if len(b) > 0:
                raise TypeError(
                    f"scbuiltin '{func.__name__}' function is not "
                    f"supported between types '{type(a).__name__}' "
                    f"and '{type(*b).__name__}'")
            else:
                raise TypeError(
                    f"scbuiltin '{func.__name__}' function is not "
                    f"supported for type '{type(a).__name__}' ")

        scbuiltin_.__name__ = func.__name__  # used to obtain special_index.
        scbuiltin_.__qualname__ += func.__name__
        return scbuiltin_

    @staticmethod
    def narop(func):
        def scbuiltin_(x, *args):
            try:
                return x._compose_narop(func, *args)
            except AttributeError:
                try:
                    return func(x, *args)
                except TypeError:
                    pass
            raise TypeError(f"scbuiltin '{func.__name__}' function is not "
                            f"supported for type '{type(x).__name__}' "
                            f"with parameteres {args}")

        scbuiltin_.__name__ = func.__name__  # used to obtain special_index.
        scbuiltin_.__qualname__ += func.__name__
        return scbuiltin_


### Random ###

# Have special index.

@scbuiltin.unop
def rand(x):
    # *** TODO: See the actual implementations.
    if type(x) is float:
        return _libsc3.main._rgen.random() * x
    elif type(x) is int:
        if x >= 1:
            return _libsc3.main._rgen.randrange(0, x, 1)
        elif x <= -1:
            return _libsc3.main._rgen.randrange(0, x, -1)
        else:
            return 0
    raise TypeError

@scbuiltin.unop
def rand2(x):
    if type(x) is float:
        return _libsc3.main._rgen.random() * x * 2 - x  # random.uniform
    elif type(x) is int:
        if x >= 0:
            return _libsc3.main._rgen.randint(-x, x)
        else:
            return _libsc3.main._rgen.randint(x, -x)
    raise TypeError

@scbuiltin.unop
def linrand(x):
    if type(x) is float:
        a = _libsc3.main._rgen.random()
        b = _libsc3.main._rgen.random()
        return min(a, b) * x
    elif type(x) is int:
        if x >= 1:
            a = _libsc3.main._rgen.randrange(0, x, 1)
            b = _libsc3.main._rgen.randrange(0, x, 1)
            return min(a, b)
        elif x <= -1:
            a = _libsc3.main._rgen.randrange(0, x, -1)
            b = _libsc3.main._rgen.randrange(0, x, -1)
            return min(a, b)
        else:
            return 0
    raise TypeError

@scbuiltin.unop
def bilinrand(x):
    if type(x) is float:
        a = _libsc3.main._rgen.random()
        b = _libsc3.main._rgen.random()
        return a - b
    elif type(x) is int:
        if x >= 1:
            a = _libsc3.main._rgen.randrange(0, x, 1)
            b = _libsc3.main._rgen.randrange(0, x, 1)
            return a - b
        elif x <= -1:
            a = _libsc3.main._rgen.randrange(0, x, -1)
            b = _libsc3.main._rgen.randrange(0, x, -1)
            return a - b
        else:
            return 0
    raise TypeError

@scbuiltin.unop
def sum3rand(x):
    # // Larry Polansky's poor man's gaussian generator.
    return (
        (_libsc3.main._rgen.random() +
        _libsc3.main._rgen.random() +
        _libsc3.main._rgen.random() - 1.5) *
        0.666666667 * x)

@scbuiltin.unop
def coin(x):
    if type(x) is float:  # sclang Float method.
        return _libsc3.main._rgen.random() < x
    elif type(x) is int:  # sclang SimpleNumber behaviour.
        if x == 0:
            return False
        else:
            return True
    raise TypeError

@scbuiltin.binop
def rrand(a, b):
    if type(a) is float or type(b) is float:
        return a + _libsc3.main._rgen.random() * (b - a)
    elif type(a) is type(b) is int:
        if a <= b:
            if b - a >= 1:
                return _libsc3.main._rgen.randrange(a, b, 1)
            else:
                return a
        else:
            return _libsc3.main._rgen.randrange(a, b, -1)
    raise TypeError

@scbuiltin.binop
def exprand(a, b):  # exprandrng
    return a * exp(log(b / a) * _libsc3.main._rgen.random())

# Don't have special index.

@scbuiltin.binop
def xrand(x, exclude=0):
    if type(x) is type(exclude) is int:
        return mod(exclude + rand(x - 1) + 1, x);
    raise TypeError

@scbuiltin.binop
def xrand2(x, exclude=0):
    if type(x) is float:
        return rand2(x)
    elif type(x) is type(exclude) is int:
        res = rand(x * 2) - x
        return x if res == exclude else res
    raise TypeError


### Unary ###

# // this is a function for preventing pathological math operations in ugens.
# // can be used at the end of a block to fix any recirculating filter values.
@scbuiltin.unop
def zapgremlins(x):
    # float32 absx = std::abs(x);
    # // very small numbers fail the first test, eliminating denormalized numbers
    # //    (zero also fails the first test, but that is OK since it returns zero.)
    # // very large numbers fail the second test, eliminating infinities
    # // Not-a-Numbers fail both tests and are eliminated.
    # return (absx > (float32)1e-15 && absx < (float32)1e15) ? x : (float32)0.;
    absx = math.abs(x)
    if absx > 1e-15 and absx < 1e15: return x
    return 0.

@scbuiltin.unop
def log2(x):
    if x == 0: return float('-inf')
    return math.log2(x)

@scbuiltin.unop
def log10(x):
    if x == 0: return float('-inf')
    return math.log10(x)

@scbuiltin.unop
def log(x):  #, base=math.e):  # In SI is unary, can't change the base.
    if x == 0: return float('-inf')
    return math.log(x)

@scbuiltin.unop
def exp(x):
    return math.exp(x)

@scbuiltin.unop
def sin(x):
    return math.sin(x)

@scbuiltin.unop
def cos(x):
    return math.cos(x)

@scbuiltin.unop
def tan(x):
    return math.tan(x)

@scbuiltin.unop
def asin(x):
    return math.asin(x)

@scbuiltin.unop
def acos(x):
    return math.acos(x)

@scbuiltin.unop
def atan(x):
    return math.atan(x)

@scbuiltin.unop
def sinh(x):
    return math.sinh(x)

@scbuiltin.unop
def cosh(x):
    return math.cosh(x)

@scbuiltin.unop
def tanh(x):
    return math.tanh(x)

# Python math module.
# @scbuiltin.unop
# def log1p(x):
#     return math.log1p(x)
# @scbuiltin.unop
# def asinh(x):
#     return math.asinh(x)
# @scbuiltin.unop
# def acosh(x):
#     return math.acosh(x)
# @scbuiltin.unop
# def atanh(x):
#     return math.atanh(x)

_ONETWELFTH = 1. / 12.
_ONE440TH = 1. / 440.

@scbuiltin.unop
def midicps(note):
    # return (float64)440. * std::pow((float64)2., (note - (float64)69.) * (float64)0.08333333333333333333333333);
    return 440. * pow(2., (note - 69.) * _ONETWELFTH)

@scbuiltin.unop
def cpsmidi(freq):
    # return sc_log2(freq * (float64)0.002272727272727272727272727) * (float64)12. + (float64)69.;
    return log2(freq * _ONE440TH) * 12. + 69.

@scbuiltin.unop
def midiratio(midi):
    #return std::pow((float32)2. , midi * (float32)0.083333333333);
    return pow(2., midi * _ONETWELFTH)

@scbuiltin.unop
def ratiomidi(ratio):
    #return (float32)12. * sc_log2(ratio);
    return 12. * log2(ratio)

@scbuiltin.unop
def octcps(note):
    # return (float32)440. * std::pow((float32)2., note - (float32)4.75);
    return 440. * pow(2., note - 4.75)

@scbuiltin.unop
def cpsoct(freq):
    # return sc_log2(freq * (float32)0.0022727272727) + (float32)4.75;
    return log2(freq * _ONE440TH + 4.75)

@scbuiltin.unop
def ampdb(amp):
    # return std::log10(amp) * (float32)20.;
    return log10(amp) * 20.

@scbuiltin.unop
def dbamp(db):
    # return std::pow((float32)10., db * (float32).05);
    return pow(10., db * .05)

@scbuiltin.unop
def squared(x):
    return x * x;

@scbuiltin.unop
def cubed(x):
    return x * x * x;

@scbuiltin.unop
def sqrt(x):
    if x < 0.:
        return -math.sqrt(-x)
    else:
        return math.sqrt(x)

@scbuiltin.unop
def hanwindow(x):
    # if (x < (float32)0. || x > (float32)1.) return (float32)0.;
    # return (float32)0.5 - (float32)0.5 * static_cast<float32>(cos(x * (float32)twopi));
    if x < 0. or x > 1.: return 0.
    return 0.5 - 0.5 * cos(x * twopi)

@scbuiltin.unop
def welwindow(x):
    # if (x < (float32)0. || x > (float32)1.) return (float32)0.;
    # return static_cast<float32>(sin(x * pi));
    if x < 0. or x > 1.: return 0.
    return sin(x * pi)

@scbuiltin.unop
def triwindow(x):
    # if (x < (float32)0. || x > (float32)1.) return (float32)0.;
    # if (x < (float32)0.5) return (float32)2. * x;
    # else return (float32)-2. * x + (float32)2.;
    if x < 0. or x > 1.: return 0.
    if x < 0.5: return 2. * x
    return -2. * x + 2.

@scbuiltin.unop
def bitriwindow(x):  # not used in sclang
    # float32 ax = (float32)1. - std::abs(x);
    # if (ax <= (float32)0.) return (float32)0.;
    # return ax;
    ax = 1. - abs(x)
    if ax <= 0.: return 0.
    return ax

@scbuiltin.unop
def rectwindow(x):
    # if (x < (float32)0. || x > (float32)1.) return (float32)0.;
    # return (float32)1.;
    if x < 0. or x > 1.: return 0.
    return 1.

@scbuiltin.unop
def scurve(x):
    # if (x <= (float32)0.) return (float32)0.;
    # if (x >= (float32)1.) return (float32)1.;
    # return x * x * ((float32)3. - (float32)2. * x);
    if x <= 0.: return 0.
    if x >= 1.: return 1.
    return x * x * (3. - 2. * x)

@scbuiltin.unop
def scurve0(x):  # not used in sclang
    # // assumes that x is in range
    # return x * x * ((float32)3. - (float32)2. * x);
    return x * x * (3. - 2. * x)

@scbuiltin.unop
def ramp(x):
    # if (x <= (float32)0.) return (float32)0.;
    # if (x >= (float32)1.) return (float32)1.;
    # return x;
    if x <= 0.: return 0.
    if x >= 1.: return 1.
    return x

@scbuiltin.unop
def sign(x):
    # return x < (float32)0. ? (float32)-1. : (x > (float32)0. ? (float32)1.f : (float32)0.f);
    if x < 0.: return -1.
    if x > 0.: return 1.
    return 0.

@scbuiltin.unop
def distort(x):
    # return x / ((float32)1. + std::abs(x));
    return x / (1. + abs(x))

@scbuiltin.unop
def distortneg(x):  # not used in sclang
    # if (x < (float32)0.)
    #     return x/((float32)1. - x);
    # else
    #     return x;
    if x < 0.: return x / (1. - x)
    return x

@scbuiltin.unop
def softclip(x):
    # float32 absx = std::abs(x);
    # if (absx <= (float32)0.5) return x;
    # else return (absx - (float32)0.25) / x;
    absx = abs(x)
    if absx <= 0.5: return x
    return (absx - 0.25) / x

@scbuiltin.unop
def even(x):
    return int(x) & 1 == 0

@scbuiltin.unop
def odd(x):
    return int(x) & 1 == 1

@scbuiltin.unop
def taylorsin(x):
    # // Taylor expansion out to x**9/9! factored  into multiply-adds
    # // from Phil Burk.
    # // valid range from -pi/2 to +3pi/2
    # x = static_cast<float32>((float32)pi2 - std::abs(pi2 - x));
    # float32 x2 = x * x;
    # return static_cast<float32>(x*(x2*(x2*(x2*(x2*(1.0/362880.0)
    #         - (1.0/5040.0))
    #         + (1.0/120.0))
    #         - (1.0/6.0))
    #         + 1.0));
    x = pi2 - abs(pi2 - x)
    x2 = x * x
    return (x*(x2*(x2*(x2*(x2*(1.0/362880.0)\
        - (1.0/5040.0))\
        + (1.0/120.0))\
        - (1.0/6.0))\
        + 1.0))

# @scbuiltin.unop
# def trunc(x):  # duplicated as unary and binary, AbstractFunction use binary, see below.
#     return math.trunc(x);

@scbuiltin.unop
def ceil(x):
    return math.ceil(x)

@scbuiltin.unop
def floor(x):
    return math.floor(x)

@scbuiltin.unop
def reciprocal(x):
    return 1. / x

@scbuiltin.unop
def bitnot(x):
    #return (float32) ~ (int)x;
    return float(~int(x))

@scbuiltin.unop
def frac(x):
    # return x - sc_floor(x);
    # return math.fmod(x)[0]
    return x - floor(x)

_ONESIXTH = 1. / 6.

@scbuiltin.narop
def lg3interp(x1, a, b, c, d): # sc_lg3interp solo la define para float32
    # // cubic lagrange interpolator
    # float32 x0 = x1 + 1.f;
    # float32 x2 = x1 - 1.f;
    # float32 x3 = x1 - 2.f;
    #
    # float32 x03 = x0 * x3 * 0.5f;
    # float32 x12 = x1 * x2 * 0.16666666666666667f;
    #
    # return x12 * (d * x0 - a * x3) + x03 * (b * x2 - c * x1);
    x0 = x1 + 1.
    x2 = x1 - 1.
    x3 = x1 - 2.
    x03 = x0 * x3 * 0.5
    x12 = x1 * x2 * _ONESIXTH
    return x12 * (d * x0 - a * x3) + x03 * (b * x2 - c * x1)

@scbuiltin.binop
def calcfeedback(delaytime, decaytime):  # CalcFeedback, solo la define para float32
    # if (delaytime == 0.f || decaytime == 0.f)
    #     return 0.f;
    #
    # float32 absret = static_cast<float32>( std::exp(log001 * delaytime / std::abs(decaytime)));
    # float32 ret = std::copysign(absret, decaytime);
    # return ret;
    if delaytime == 0. or decaytime == 0.:
        return 0.
    absret = math.exp(log001 * delaytime / abs(decaytime))
    return math.copysign(absret, decaytime)

@scbuiltin.unop
def wrap1(x):
    # if (x >= (float32) 1.) return x + (float32)-2.;
    # if (x <  (float32)-1.) return x + (float32) 2.;
    # return x;
    if x >= 1.: return x - 2.
    if x < -1.: return x + 2.
    return x

@scbuiltin.unop
def fold1(x):
    # if (x >= (float32) 1.) return (float32) 2. - x;
    # if (x <  (float32)-1.) return (float32)-2. - x;
    # return x;
    if x >= 1.: return 2 - x
    if x < -1.: return -2 - x
    return x

@scbuiltin.unop
def graycode(x): # grayCode, está abajo de todo y es para int32
    # return x ^ (x >> 1);
    x = int(x)
    return x ^ (x >> 1)

@scbuiltin.unop
def degrad(x):
    return x * pi / 180.

@scbuiltin.unop
def raddeg(x):
    return x * 180. / pi

@scbuiltin.unop
def next_power_of_two(x):  # Integer _NextPowerOfTwo
    return 2 ** ceil(log(x) / log(2))

@scbuiltin.binop
def next_near_power(x, base=2):  # SimpleNumber.nextPowerOf
    return base ** ceil(log(x) / log(base))

@scbuiltin.binop
def previous_near_power(x, base=2):  # SimpleNumber.previousPowerOf
    return base ** (ceil(log(x) / log(base)) - 1)


### Binary ###

@scbuiltin.binop
def mod(a, b):
    # NOTE: Has a different behaviour from Python's %.
    # DOUBLE/FLOAT
    # // avoid the divide if possible
    # const double lo = (double)0.;
    # if (in >= hi) {
    #     in -= hi;
    #     if (in < hi) return in;
    # } else if (in < lo) {
    #     in += hi;
    #     if (in >= lo) return in;
    # } else return in;
    #
    # if (hi == lo) return lo;
    # FLOAT
    # return in - hi*sc_floor(in/hi);
    # INT
    # int c;
    # c = in % hi;
    # if (c<0) c += hi;
    # return c;
    zero = 0
    if a >= b:
        a -= b
        if a < b: return a
    elif a < zero:
        a += b
        if a >= zero: return a
    else:
        return a

    if b == zero: return zero

    if type(a) is float or type(b) is float:
        return a - b * floor(a/b)

    c = int(math.fmod(a, b))
    if c < 0: c += b
    return c

@scbuiltin.narop
def wrap(x, lo, hi, range=None): # *** BUG: AbstractFunction usa sin range. tiene dos firmas, sin y con range, la implementación varía sutilmente.
# INT: abajo define wrap para int sin range como:
# return sc_mod(in - lo, hi - lo + 1) + lo;
    if type(x) is int:
        lo = int(lo)
        hi = int(hi)
        return mod(x - lo, hi - lo + 1) + lo
    # // avoid the divide if possible
    # if (in >= hi) {
# range = hi - lo; # sin range
    #     in -= range;
    #     if (in < hi) return in;
    # } else if (in < lo) {
# range = hi - lo; # sin range
    #     in += range;
    #     if (in >= lo) return in;
    # } else return in;
    #
    # if (hi == lo) return lo;
    # return in - range * sc_floor((in - lo)/range);
    if x >= hi:
        if range is None:
            range = hi - lo
        x -= range
        if x < hi: return x
    elif x < lo:
        if range is None:
            range = hi - lo
        x += range
        if x >= lo: return x
    else:
        return x

    if hi == lo: return lo
    return x - range * floor((x - lo) / range)

@scbuiltin.narop
def fold(x, lo, hi, range=None, range2=None): # *** BUG: ídem wrap con range y range2
# INT: abajo define fold para int sin range ni range2
# int b = hi - lo;
# int b2 = b+b;
# int c = sc_mod(in - lo, b2);
# if (c>b) c = b2-c;
# return c + lo;
    if type(x) is int:
        lo = int(lo)
        hi = int(hi)
        b = hi - lo
        b2 = b + b
        c = mod(x - lo, b2)
        if c > b: c = b2 - c
        return c + lo
    # double x, c;
    # x = in - lo;
    #
    # // avoid the divide if possible
    # if (in >= hi) {
    #     in = hi + hi - in;
    #     if (in >= lo) return in;
    # } else if (in < lo) {
    #     in = lo + lo - in;
    #     if (in < hi) return in;
    # } else return in;
    #
    # if (hi == lo) return lo;
    # // ok do the divide
# range = hi - lo; # sin range
# range2 = range + range; # sin range2
    # c = x - range2 * sc_floor(x / range2);
    # if (c>=range) c = range2 - c;
    # return c + lo;
    x2 = x - lo
    if x >= hi:
        x = hi + hi - x
        if x >= lo: return x
    elif x < lo:
        x = lo + lo - x
        if x < hi: return x
    else:
        return x

    if hi == lo: return lo
    if range is None:
        range = hi - lo
        range2 = range + range
    c = x2 - range2 * floor(x2 / range2)
    if c >= range:
        c = range2 - c
    return c + lo

# TODO: define pow, eso no es bueno para mi. *********************************
# @scbuiltin
# def pow(a, b):
#     # return a >= 0.f ? std::pow(a, b) : -std::pow(-a, b);
#     pass

@scbuiltin.binop
def div(a, b): # TODO: define div para int devolviendo el dividendo si el divisor es cero, en sclang es el comportamiento de 1 div: 0, en Python 1 // 0 es error.
               # TODO: ver si se usa para las ugens o qué cómo, lo mismo con mod.
               # TODO: si lo sargumentos son float sclang realiza las operaciones y castead el valor de retorno.
    # int c;
    # if (b) {
    #     if (a<0) c = (a+1)/b - 1;
    #     else c = a/b;
    # } else c = a;
    # return c;
    if b:
        if a < 0: c = (a + 1) / b - 1
        else: c = a / b
    else:
        c = a
    return int(c)

@scbuiltin.binop
def min(a, b):
    return builtins.min(a, b)

@scbuiltin.binop
def max(a, b):
    return builtins.max(a, b)

@scbuiltin.binop
def round(x, quant=1):
    # return quant==0. ? x : sc_floor(x/quant + .5) * quant;
    # INT return quant==0 ? x : sc_div(x + quant/2, quant) * quant;
    if type(x) is int:
        quant = int(quant)
        if quant == 0:
            return float(x)
        else:
            return float(div(x + quant // 2, quant) * quant)
    if quant == 0.:
        return float(x)
    else:
        return float(floor(x / quant + .5) * quant)

@scbuiltin.binop
def roundup(x, quant=1):
    # return quant==0. ? x : sc_ceil(x/quant) * quant;
    # INT return quant==0 ? x : sc_div(x + quant - 1, quant) * quant;
    if type(x) is int:
        quant = int(quant)
        if quant == 0:
            return float(x)
        else:
            return float(div(x + quant - 1, quant) * quant)
    if quant == 0.:
        return float(x)
    else:
        return float(ceil(x / quant) * quant)

@scbuiltin.binop
def trunc(x, quant=1):
    # return quant==0. ? x : sc_floor(x/quant) * quant;
    # INT: return quant==0 ? x : sc_div(x, quant) * quant;
    if type(x) is int:
        quant = int(quant)
        if quant == 0:
            return float(x)
        else:
            return float(div(x, quant) * quant)
    if quant == 0.:
        return float(x)
    else:
        return float(floor(x / quant) * quant)

@scbuiltin.binop
def atan2(a, b): # TODO: Solo la define para float. Pero creo que no define sin/cos y las demás.
    # return std::atan2(a, b);
    return math.atan2(a, b)

_SQRT2M1 = math.sqrt(2.) - 1.;

@scbuiltin.narop
def clip(x, lo, hi):
    #./common/SC_BoundsMacros.h:
    #inline T sc_clip(T x, U lo, V hi) return std::max(std::min(x, (T)hi), (T)lo);
    T = type(x)
    return max(min(x, T(hi)), T(lo))

@scbuiltin.binop
def hypot(x, y):
    return math.hypot(x, y)

@scbuiltin.binop
def hypotx(x, y):  # hypotenuse aproximation C name, hypotApx in sclang.
    # double minxy;
    # x = std::abs(x);
    # y = std::abs(y);
    # minxy = sc_min(x,y);
    # return x + y - kDSQRT2M1 * minxy;
    x = abs(x)
    y = abs(y)
    minxy = min(x, y)
    return x + y - _SQRT2M1 * minxy

@scbuiltin.binop
def gcd(a, b):
# FLOAT: abajo define para float gcd(u, v)
# return (float) sc_gcd((long) std::trunc(u), (long) std::trunc(v));
    # if (a == 0)
    #     return b;
    # if (b == 0)
    #     return a;
    # const bool negative = (a <= 0 && b <= 0);
    # a = sc_abs(a);
    # b = sc_abs(b);
    # if (a == 1 || b == 1) {
    #     if(negative) {
    #         return (long) -1;
    #     } else {
    #         return (long) 1;
    #     }
    # }
    # if (a < b) {
    #     long t = a;
    #     a = b;
    #     b = t;
    # }
    # while (b > 0) {
    #     long t = a % b;
    #     a = b;
    #     b = t;
    # }
    # if(negative) {
    #     a = 0 - a;
    # }
    # return a;
    if type(a) is float or type(b) is float:
        a = int(a)
        b = int(b)
    T = type(a)

    if a == 0: return b
    if b == 0: return a
    negative = a <= 0 and b <=0
    a = abs(a)
    b = abs(b)
    if a == 1 or b == 1:
        if negative: return -1
        else: return 1
    if a < b:
        t = a
        a = b
        b = t
    while b > 0:
        t = a % b
        a = b
        b = t
    if negative:
        a = 0 - a

    return T(a)

@scbuiltin.binop
def lcm(a, b):
# FLOAT: abajo define para float lcm(u, v)
# return (float) sc_lcm((long) std::trunc(u), (long) std::trunc(v));
    # if (a == 0 || b == 0)
    #     return (long)0;
    # else
    #     return (a * b) / sc_gcd(a, b);
    if type(a) is float or type(b) is float:
        a = int(a)
        b = int(b)

    if a == 0 or b == 0:
        return 0
    else:
        return (a * b) // gcd(a, b)

# @scbuiltin.binop
# def bitand(a, b):  # has special index
#     return a & b
#
# @scbuiltin.binop
# def bitor(a, b):  # has special index
#     return a | b

# bitxor  # missing
# bitHammingDistance  # missing

# @scbuiltin.binop
# def leftshift(a, b):  # has special index
#     return a << b
#
# @scbuiltin.binop
# def rightshift(a, b):  # has special index
#     return a >> b
#
# @scbuiltin.binop
# def urightshift(a, b):  # has special index
#     return (uint32)a >> b;

# @scbuiltin.binop # only used in /server/plugins/LFUGens.cpp and example in ServerPluginAPI.schelp
# def powi(x, n):
#     # F z = 1;
#     # while (n != 0)
#     # {
#     #     if ((n & 1) != 0)
#     #     {
#     #         z *= x;
#     #     }
#     #     n >>= 1;
#     #     x *= x;
#     # }
#     # return z;
#     pass

@scbuiltin.binop
def thresh(a, b): # sc_thresh(T a, U b)
    # return a < b ? (T)0 : a;
    T = type(a)
    if a < b: return T(0)
    return a

@scbuiltin.binop
def clip2(a, b): # sc_clip2(T a, T b) # Estas son las que usa como operadores integrados de las UGens.
    # return sc_clip(a, -b, b);
    return clip(a, -b, b)

@scbuiltin.binop
def wrap2(a, b): # wrap2(T a, T b)
    # return sc_wrap(a, -b, b);
    return wrap(a, -b, b)

@scbuiltin.binop
def fold2(a, b): # sc_fold2(T a, T b)
    # return sc_fold(a, -b, b);
    return fold(a, -b, b)

@scbuiltin.binop
def excess(a, b): # sc_excess(T a, T b)
    # return a - sc_clip(a, -b, b);
    return a - clip(a, -b, b)

@scbuiltin.binop
def first_arg(a, b):
    return a

@scbuiltin.binop
def scaleneg(a, b):
    # template: T a, T b
    # if (a < 0)
    #     return a*b;
    # else
    #     return a;
    # DOUBLE/FLOAT
    # b = 0.5 * b + 0.5;
    # return (std::abs(a) - a) * b + a;
    if type(a) is int and type(b) is int:
        if a < 0: return a * b
        return a
    b = 0.5 * b + 0.5
    return (abs(a) - a) * b + a

@scbuiltin.binop
def amclip(a, b):
    # template: T a, T b
    # if (b < 0)
    #     return 0;
    # else
    #     return a*b;
    # DOUBLE/FLOAT
    # return a * 0.5 * (b + std::abs(b));
    if type(a) is int and type(b) is int:
        if b < 0: return 0
        return a * b
    return a * 0.5 * (b + abs(b))

@scbuiltin.binop
def ring1(a, b):
    return a * b + a

@scbuiltin.binop
def ring2(a, b):
    return a * b + a + b

@scbuiltin.binop
def ring3(a, b):
    return a * a * b

@scbuiltin.binop
def ring4(a, b):
    return a * a * b - a * b * b

@scbuiltin.binop
def difsqr(a, b):
    return a * a - b * b

@scbuiltin.binop
def sumsqr(a, b):
    return a * a + b * b

@scbuiltin.binop
def sqrsum(a, b):
    z = a + b
    return z * z

@scbuiltin.binop
def sqrdif(a, b):
    z = a - b
    return z * z

@scbuiltin.binop
def absdif(a, b):
    return abs(a - b)


### Nary ###

@scbuiltin.narop
def blend(a, b, frac=0.5):
    # // frac should be from zero to one
    return a + (frac * (b - a))

@scbuiltin.narop
def snap(x, resolution=1.0, margin=0.05, strength=1.0):
    round_ = round(x, resolution)
    diff = round_ - x
    if abs(diff) < margin:
        return x + strength * diff
    else:
        return x

@scbuiltin.narop
def softround(x, resolution=1.0, margin=0.05, strength=1.0):
    round_ = round(x, resolution)
    diff = round_ - x
    if abs(diff) > margin:
        return x + strength * diff
    else:
        return x

@scbuiltin.narop
def linlin(x, inmin, inmax, outmin, outmax, clip='minmax'):
    if clip == 'minmax':
        if x <= inmin: return outmin
        if x >= inmax: return outmax
    elif clip == 'min':
        if x <= inmin: return outmin
    elif clip == 'max':
        if x >= inmax: return outmax
    return (x - inmin) / (inmax - inmin) * (outmax - outmin) + outmin

@scbuiltin.narop
def linexp(x, inmin, inmax, outmin, outmax, clip='minmax'):
    if clip == 'minmax':
        if x <= inmin: return outmin
        if x >= inmax: return outmax
    elif clip == 'min':
        if x <= inmin: return outmin
    elif clip == 'max':
        if x >= inmax: return outmax
    return math.pow(outmax / outmin, (x - inmin) / (inmax - inmin)) * outmin

@scbuiltin.narop
def explin(x, inmin, inmax, outmin, outmax, clip='minmax'):
    if clip == 'minmax':
        if x <= inmin: return outmin
        if x >= inmax: return outmax
    elif clip == 'min':
        if x <= inmin: return outmin
    elif clip == 'max':
        if x >= inmax: return outmax
    return log(x / inmin, math.e) / log(inmax / inmin, math.e)\
           * (outmax - outmin) + outmin

@scbuiltin.narop
def expexp(x, inmin, inmax, outmin, outmax, clip='minmax'):
    if clip == 'minmax':
        if x <= inmin: return outmin
        if x >= inmax: return outmax
    elif clip == 'min':
        if x <= inmin: return outmin
    elif clip == 'max':
        if x >= inmax: return outmax
    return math.pow(
        outmax / outmin,
        log(x / inmin, math.e) / log(inmax / inmin, math.e)
    ) * outmin

@scbuiltin.narop
def lincurve(x, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
    if clip == 'minmax':
        if x <= inmin: return outmin
        if x >= inmax: return outmax
    elif clip == 'min':
        if x <= inmin: return outmin
    elif clip == 'max':
        if x >= inmax: return outmax
    if abs(curve) < 0.001:
        # // If the value should be clipped, it has already been clipped (above).
        # // If we got this far, then linlin does not need to do any clipping.
        # // Inlining the formula here makes it even faster.
        return (x - inmin) / (inmax - inmin) * (outmax - outmin) + outmin
    grow = math.exp(curve)
    a = (outmax - outmin) / (1.0 - grow)
    b = outmin + a
    scaled = (x - inmin) / (inmax - inmin)
    return b - a * math.pow(grow, scaled, math.e)

@scbuiltin.narop
def curvelin(x, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
    if clip == 'minmax':
        if x <= inmin: return outmin
        if x >= inmax: return outmax
    elif clip == 'min':
        if x <= inmin: return outmin
    elif clip == 'max':
        if x >= inmax: return outmax
    if abs(curve):
        # // If the value should be clipped, it has already been clipped (above).
        return (x - inmin) / (inmax - inmin) * (outmax - outmin) + outmin
    grow = math.exp(curve)
    a = (inmax - inmin) / (1.0 - grow)
    b = inmin + a
    return log((b - x) / a, math.e) * (outmax - outmin) / curve + outmin

@scbuiltin.narop
def bilin(x, incenter, inmin, inmax, outcenter, outmin, outmax, clip='minmax'):
    if clip == 'minmax':
        if x <= inmin: return outmin
        if x >= inmax: return outmax
    elif clip == 'min':
        if x <= inmin: return outmin
    elif clip == 'max':
        if x >= inmax: return outmax
    if x >= incenter:
        return linlin(x, incenter, inmax, outcenter, outmax, None)
    else:
        return linlin(inmin, incenter, outmin, outcenter, None)

@scbuiltin.narop
def biexp(x, incenter, inmin, inmax, outcenter, outmin, outmax, clip='minmax'):
    if clip == 'minmax':
        if x <= inmin: return outmin
        if x >= inmax: return outmax
    elif clip == 'min':
        if x <= inmin: return outmin
    elif clip == 'max':
        if x >= inmax: return outmax
    if x >= incenter:
        return explin(incenter, inmax, outcenter, outmax, None)
    else:
        return explin(inmin, incenter, outmin, outcenter, None)

@scbuiltin.narop
def moddif(a, b, mod=1.0):
    diff = mod(absdif(a, b), mod)
    modhalf = mod * 0.5
    return modhalf - absdif(diff, modhalf)

@scbuiltin.narop
def lcurve(x, a=1.0, m=0.0, n=1.0, tau=1.0):
    x = -x
    if tau == 1.0:
        return a * (m * exp(x) + 1) / (n * exp(x) + 1)
    else:
        rtau = 1. / tau
        return a * (m * exp(x) * rtau + 1) / (n * exp(x) * rtau + 1)

@scbuiltin.narop  # binop
def gauss(x, stdev):  # standard_deviation
    return sqrt(-2 * log(rand(1.0))) * sin(rand(twopi)) * stdev + x

@scbuiltin.narop
def gauss_curve(x, a=1.0, b=0.0, c=1.0):
    return a * exp(squared(x - b) / (-2.0 * squared(c)))

# digitValue  # ?
# thru ??  # ?

# complex and cartesian
# rho
# theta
# rotate
# dist


### Array ###

# The next sclang functios operate over lists and don't map to server opcodes.
# They are Routine's state aware, i.e. random operatons use rgen state.

import operator
import itertools


def normalize(lst, min=0.0, max=1.0):
    min = itertools.repeat(min)
    max = itertools.repeat(max)
    outmin = itertools.repeat(builtins.min(lst))
    outmax = itertools.repeat(builtins.max(lst))
    return list(map(linlin, lst, outmin, outmax, min, max))

def normsum(lst):  # normalizeSum/normalize_sum
    return list(map(operator.truediv, lst, itertools.repeat(sum(lst))))

def shuffle(lst, random=None):  # scramble
    return _libsc3.main._rgen.shuffle(lst, random)  # In place.

# mirror, mirror1, mirror2  # one mirror with mode.
# stutter, rotate, pyramid, pyramidg, sputter(rand), etc.

def blend_at(lst, index):
    imin = int(roundup(index)) - 1
    imax = len(lst) - 1
    a = lst[clip(imin, 0, imax)]
    b = lst[clip(imin + 1, 0, imax)]
    return blend(a, b, abs(index - imin))

def resamp0(lst, new_size):
    factor = (len(lst) - 1) / builtins.max(new_size - 1, 1)
    return list(lst[int(round(i * factor))] for i in range(new_size))

def resamp1(lst, new_size):
    factor = (len(lst) - 1) / builtins.max(new_size - 1, 1)
    return list(blend_at(lst, i * factor) for i in range(new_size))

def index_of_greater_than(lst, val, start=0):
    for i, n in enumerate(lst[start:]):
        if n > val:
            return i + start
    return len(lst) - 1

def index_in_between(lst, val):
    # // Collection is sorted, returns linearly interpolated index.
    i = index_of_greater_than(lst, val)
    if i == 0:
        return 0
    if i == len(lst):
        return len(lst) - 1
    a = lst[i - 1]
    b = lst[i]
    div = b - a
    if div == 0:
        return i
    return ((val - a) / div) + i - 1

def as_random_table(lst, size=None):
    if size is None:
        size = len(lst)
    else:
        lst = resamp1(lst, size)
    acc = list(itertools.accumulate(lst))  # // Incrementally integrate.
    norm = normalize(acc, 0.0, size - 1.0)  # // Normalize and scale by max index.
    lasti = 0
    res = []
    for i in range(size):
        # indexInBetween wit start and rescale to 0..1
        lasti = index_of_greater_than(norm, i, lasti)
        a = norm[lasti-1]
        div = norm[lasti] - a
        if div == 0:
            res.append(lasti / size)
        else:
            res.append( (((i - a) / div) + lasti - 1) / size )
    return res

def table_rand(lst):
    return blend_at(lst, rand(float(len(lst) - 1)))


### SequenceableCollection ###

def choice(lst):  # choose
    return _libsc3.main._rgen.choice(lst)

def choices(lst, weights=None, *, cum_weights=None, k=1):  # wchoose
    return _libsc3.main._rgen.choices(
        lst, weights, cum_weights=cum_weights, k=k)

# ...


### Utilities ###


_uid_counter = itertools.count()

def uid():
    '''Library wise counter used to get unique ids for server replies.'''
    return next(_uid_counter)


def counter(stop):
    '''Possibly infinite counter, range replacement for patterns.'''
    if stop == float('inf'):
        return itertools.count()
    else:
        return range(stop)



"""SystemSynthdefs.sc // synthdefs needed by classes."""

from ..base import classlibrary as clb
from ..base import systemactions as sac
from . import synthdef as sdf
from . import ugens as ugns
from . import envelope as evp


__all__ = ['SystemDefs']


class MetaSystemDefs(type):
    CHANNELS = 16
    PREFIX = 'temp__'
    MAX_TMP_DEF_NAMES = 512
    _tmp_def_count = 0
    _sdefs = dict()

    def __init__(cls, *_):
        clb.ClassLibrary.add(cls, lambda _: cls._build())

    @property
    def synthdefs(cls):
        return list(cls._sdefs.values())

    def generate_tmp_name(cls):
        name = cls.PREFIX + str(cls._tmp_def_count)
        cls._tmp_def_count += 1
        cls._tmp_def_count %= cls.MAX_TMP_DEF_NAMES
        return name

    def add_synthdef(cls, name):
        sdef = cls._sdefs[name]
        sdef.add()  # Running servers or offline patterns.
        sac.ServerBoot.add('all', lambda server: sdef.add())  # Next boot.

    def add_all(cls):
        for sdef in cls._sdefs.values():
            sdef.add()  # Running servers or offline patterns.
            sac.ServerBoot.add('all', lambda server: sdef.add())  # Next boot.

    def _build(cls):
        def default(freq=220, index=2.5, fmh=1, amp=0.1, pan=0, gate=1):
            fc = freq
            fm = fc * fmh.lag()
            d = index.lag() * fm
            mod = ugns.SinOsc.ar(fm) * d
            car = ugns.SinOsc.ar(fc + mod) * amp.lag()
            res = car * ugns.EnvGen.kr(
                evp.Env.adsr(0.05, 0.1, 0.8, 0.1), gate, done_action=2)
            res = ugns.LPF.ar(res, freq * 3)
            ugns.Out.ar(0, ugns.Pan2.ar(res, pan))

        cls._sdefs['default'] = sdf.SynthDef('default', default)
        cls._sdefs['fm'] = sdf.SynthDef('fm', default)

        def test(out, amp=0.1, gate=1):
            sig = ugns.PinkNoise.ar() * amp
            env = ugns.EnvGen.kr(
                evp.Env.asr(0.01, 1, 0.01), gate, done_action=2)
            ugns.Out(out, sig * env)

        cls._sdefs['test'] = sdf.SynthDef('test', test)

        # There are other defs in not used by now.


class SystemDefs(metaclass=MetaSystemDefs):
    pass



"""BufIO.sc"""

from . import infougens as ifu
from .. import ugen as ugn
from .. import _graphparam as gpp
from ...base import main as _libsc3
from ...base import utils as utl


class PlayBuf(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels, bufnum=0, rate=1.0, trigger=1.0,
           start_pos=0.0, loop=0.0, done_action=0):
        return cls._multi_new(
            'audio', num_channels, bufnum, rate,
            trigger, start_pos, float(loop), done_action)

    @classmethod
    def kr(cls, num_channels, bufnum=0, rate=1.0, trigger=1.0,
           start_pos=0.0, loop=0.0, done_action=0):
        return cls._multi_new(
            'control', num_channels, bufnum, rate,
            trigger, start_pos, float(loop), done_action)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


class TGrains(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels, trigger=0, bufnum=0, rate=1.0, center_pos=0,
           dur=0.1, pan=0, amp=0.1, interp=4):
        return cls._multi_new(
            'audio', num_channels, trigger, bufnum,
            rate, center_pos, dur, pan, amp, interp)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


# SimpleLoopBuf, commented ugen.


class BufRd(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels, bufnum=0, phase=0.0, loop=0.0, interpolation=2):
        return cls._multi_new(
            'audio', num_channels, bufnum, phase, float(loop), interpolation)

    @classmethod
    def kr(cls, num_channels, bufnum=0, phase=0.0, loop=0.0, interpolation=2):
        return cls._multi_new(
            'control', num_channels, bufnum, phase, float(loop), interpolation)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.

    def _check_inputs(self):  # override
        if self.rate == 'audio'\
        and gpp.ugen_param(self.inputs[1])._as_ugen_rate() != 'audio':
            return (f'phase input is not audio rate: {self.inputs[1]}'
                    f'{gpp.ugen_param(self.inputs[1])._as_ugen_rate()}')
        return self._check_valid_inputs()


class BufWr(ugn.UGen):
    @classmethod
    def ar(cls, input_list, bufnum=0, phase=0.0, loop=0.0):
        return cls._multi_new(
            'audio', bufnum, phase, float(loop), *utl.as_list(input_list))

    @classmethod
    def kr(cls, input_list, bufnum=0, phase=0.0, loop=0.0):
        return cls._multi_new(
            'control', bufnum, phase, float(loop), *utl.as_list(input_list))

    def _check_inputs(self):  # override
        if self.rate == 'audio':
            if gpp.ugen_param(self.inputs[1])._as_ugen_rate() != 'audio':
                return (f'phase input is not audio rate: {self.inputs[1]} '
                        f'{gpp.ugen_param(self.inputs[1])._as_ugen_rate()}')
            elif any(gpp.ugen_param(x)._as_ugen_rate() != 'audio'\
                     for x in self.inputs[3:]):
                return (f'input_list input is not audio rate: {self.inputs[3:]} '
                        f'{[gpp.ugen_param(x)._as_ugen_rate() for x in self.inputs[3:]]}')
        return self._check_valid_inputs()


class RecordBuf(ugn.UGen):
    @classmethod
    def ar(cls, input_list, bufnum=0, offset=0.0, rec_level=1.0, pre_level=0.0,
           run=1.0, loop=0.0, trigger=1.0, done_action=0):
        return cls._multi_new(
            'audio', bufnum, offset, rec_level, pre_level, run,
            float(loop), trigger, done_action, *utl.as_list(input_list))

    @classmethod
    def kr(cls, input_list, bufnum=0, offset=0.0, rec_level=1.0, pre_level=0.0,
           run=1.0, loop=0.0, trigger=1.0, done_action=0):
        return cls._multi_new(
            'control', bufnum, offset, rec_level, pre_level, run,
            float(loop), trigger, done_action, *utl.as_list(input_list))


class ScopeOut(ugn.UGen):
    @classmethod
    def ar(cls, input_list, bufnum=0):
        cls._multi_new('audio', bufnum, *utl.as_list(input_list))
        # return 0.0  # ScopeOut has no output.

    @classmethod
    def kr(cls, input_list, bufnum=0):
        cls._multi_new('control', bufnum, *utl.as_list(input_list))
        # return 0.0  # ScopeOut has no output.


class ScopeOut2(ugn.UGen):
    @classmethod
    def ar(cls, input_list, scope_num=0, max_frames=4096, scope_frames=None):
        if scope_frames is None:
            scope_frames = max_frames
        cls._multi_new(
            'audio', scope_num, max_frames,
            scope_frames, *utl.as_list(input_list))
        # return 0.0  # ScopeOut2 has no output.

    @classmethod
    def kr(cls, input_list, scope_num=0, max_frames=4096, scope_frames=None):
        if scope_frames is None:
            scope_frames = max_frames
        cls._multi_new(
            'control', scope_num, max_frames,
            scope_frames, *utl.as_list(input_list))
        # return 0.0  # ScopeOut2 has no output.


class Tap(ugn.UGen):
    @classmethod
    def ar(cls, bufnum=0, num_channels=1, delay_time=0.2):
        # // This depends on the session sample rate, not buffer.
        n = delay_time * ifu.SampleRate.ir().neg()
        return PlayBuf.ar(num_channels, bufnum, 1, 0, n, 1)


class LocalBuf(ugn.WidthFirstUGen):
    @classmethod
    def new(cls, num_frames=1, num_channels=1):
        return cls._multi_new('scalar', num_channels, num_frames)

    @classmethod
    def _new1(cls, rate, *args):  # override
        max_local_bufs = _libsc3.main._current_synthdef._max_local_bufs
        if max_local_bufs is None:
            max_local_bufs = MaxLocalBufs.new()
            _libsc3.main._current_synthdef._max_local_bufs = max_local_bufs
        max_local_bufs.increment()
        obj = cls._create_ugen_object(rate)
        obj._add_to_synth()
        return obj._init_ugen(*args, max_local_bufs)

    @classmethod
    def new_from(cls, lst):
        shape = utl.shape(lst)
        size = len(shape)
        if size == 0:
            raise TypeError(f'{cls.__name__} wrong type: {type(lst).__name__}')
        elif size == 1:
            shape = [1, len(lst)]
        elif size == 2:
            shape = list(shape)
        else:  # size > 2:
            raise ValueError(f'{cls.__name__} list has not the right shape')
        shape.reverse()
        buf = cls.new(*shape)
        buf.set(utl.flat(utl.flop(lst)))
        return buf

    @property
    def num_frames(self):
        return self.inputs[1]

    @property
    def num_channels(self):
        return self.inputs[0]

    def set(self, values, offset=0):
        SetBuf.new(self, utl.as_list(values), offset)

    def clear(self):
        ClearBuf.new(self)


class MaxLocalBufs(ugn.UGen):
    _default_rate = None

    @classmethod
    def new(cls):
        return cls._multi_new('scalar', 0)

    def increment(self):
        inputs = list(self._inputs)
        inputs[0] += 1
        self._inputs = tuple(inputs)


class SetBuf(ugn.WidthFirstUGen):
    @classmethod
    def new(cls, buf, values, offset=0):
        return cls._multi_new('scalar', buf, offset, len(values), *values)


class ClearBuf(ugn.WidthFirstUGen):
    @classmethod
    def new(cls, buf):
        return cls._multi_new('scalar', buf)



"""Poll.sc"""

from ...base import utils as utl
from .. import ugen as ugn
from .. import _graphparam as gpp
from . import oscillators as ocl


class Poll(ugn.UGen):
    @classmethod
    def ar(cls, trig, input, label=None, trig_id=-1):
        cls._multi_new_list(['audio', trig, input, label, trig_id])
        return input

    @classmethod
    def kr(cls, trig, input, label=None, trig_id=-1):
        cls._multi_new_list(['control', trig, input, label, trig_id])
        return input

    @classmethod
    def new(cls, trig, input, label=None, trig_id=-1):
        rate = [gpp.ugen_param(item)._as_ugen_rate() for item in utl.as_list(input)]
        rate = utl.unbubble(rate)
        cls._multi_new_list([rate, trig, input, label, trig_id])
        return input

    @classmethod
    def _new1(cls, rate, trig, input, label, trig_id):
        label = label or f'UGen({type(input).__name__})'
        label = [int(x) for x in bytes(label, 'utf-8')]  # *** TODO: sc ascii method
        # label = [x - 256 if x > 127 else x for x in bytes(label, 'utf-8')]  # sclang uses signed, works the same
        if rate == 'scalar':
            rate = 'control'
        if isinstance(trig, (int, float)):
            selector = ocl.Impulse._method_selector_for_rate(rate)
            trig = getattr(ocl.Impulse, selector)(trig, 0)
        obj = cls._create_ugen_object(rate)
        obj._add_to_synth()
        return obj._init_ugen(trig, input, trig_id, len(label), *label)

    def _check_inputs(self):  # override
        return self._check_sr_as_first_input()

    # init_ugen, same override pattern as EnvGen, I don't get it.



"""Noise.sc"""

from .. import ugen as ugn


class RandSeed(ugn.WidthFirstUGen):
    _default_rate = 'scalar'

    @classmethod
    def ar(cls, trig=0.0, seed=56789):
        cls._multi_new('audio', trig, seed)
        # return 0.0  # // RandSeed has no output.

    @classmethod
    def kr(cls, trig=0.0, seed=56789):
        cls._multi_new('control', trig, seed)
        # return 0.0  # // RandSeed has no output.

    @classmethod
    def ir(cls, trig=0.0, seed=56789):
        cls._multi_new('scalar', trig, seed)
        # return 0.0  # // RandSeed has no output.


class RandID(ugn.WidthFirstUGen):
    _default_rate = 'scalar'

    # // Choose which random number generator to use for this synth.
    @classmethod
    def kr(cls, id=0):
        cls._multi_new('control', id)
        # return 0.0  # // RandID has no output.

    @classmethod
    def ir(cls, id=0):
        cls._multi_new('scalar', id)
        # return 0.0  # // RandID has no output.


class Rand(ugn.UGen):
    # // Uniform distribution.
    _default_rate = None

    @classmethod
    def new(cls, lo=0.0, hi=1.0):
        return cls._multi_new('scalar', lo, hi)


class IRand(ugn.UGen):
    # // Uniform distribution of integers.
    _default_rate = None

    @classmethod
    def new(cls, lo=0, hi=127):
        return cls._multi_new('scalar', lo, hi)


class TRand(ugn.UGen):
    # // Uniform distribution.
    @classmethod
    def ar(cls, lo=0.0, hi=1.0, trig=0.0):
        return cls._multi_new('audio', lo, hi, trig)

    @classmethod
    def kr(cls, lo=0.0, hi=1.0, trig=0.0):
        return cls._multi_new('control', lo, hi, trig)


class TIRand(ugn.UGen):
    # // Uniform distribution of integers.
    @classmethod
    def ar(cls, lo=0, hi=127, trig=0.0):
        return cls._multi_new('audio', lo, hi, trig)

    @classmethod
    def kr(cls, lo=0, hi=127, trig=0.0):
        return cls._multi_new('control', lo, hi, trig)


class LinRand(ugn.UGen):
    _default_rate = None

    # // Linear distribution.
    # // if minmax <= 0 then skewed towards lo.
    # // else skewed towards hi.
    @classmethod
    def new(cls, lo=0.0, hi=1.0, minmax=0):
        return cls._multi_new('scalar', lo, hi, minmax)


class NRand(ugn.UGen):
    # // Sum of N uniform distributions.
    # // n = 1 : uniform distribution - same as Rand
    # // n = 2 : triangular distribution
    # // n = 3 : smooth hump
    # // as n increases, distribution converges towards gaussian.
    _default_rate = None

    @classmethod
    def new(cls, lo=0.0, hi=0.0, n=0):
        return cls._multi_new('scalar', lo, hi, n)


class ExpRand(ugn.UGen):
    # // Exponential distribution.
    _default_rate = None

    @classmethod
    def new(cls, lo=0.01, hi=1.0):
        return cls._multi_new('scalar', lo, hi)


class TExpRand(ugn.UGen):
    # // Exponential distribution.  # NOTE: sclang sais 'uniform'.
    @classmethod
    def ar(cls, lo=0.01, hi=1.0, trig=0.0):
        return cls._multi_new('audio', lo, hi, trig)

    @classmethod
    def kr(cls, lo=0.01, hi=1.0, trig=0.0):
        return cls._multi_new('control', lo, hi, trig)


class CoinGate(ugn.UGen):
    @classmethod
    def ar(cls, prob, input):
        return cls._multi_new('audio', prob, input)

    @classmethod
    def kr(cls, prob, input):
        return cls._multi_new('control', prob, input)


class TWindex(ugn.UGen):
    @classmethod
    def ar(cls, input, lst, normalize=0):
        return cls._multi_new('audio', input, normalize, *lst)

    @classmethod
    def kr(cls, input, lst, normalize=0):
        return cls._multi_new('control', input, normalize, *lst)


class WhiteNoise(ugn.UGen):
    @classmethod
    def ar(cls):
        return cls._multi_new('audio')

    @classmethod
    def kr(cls):
        return cls._multi_new('control')


class BrownNoise(WhiteNoise):
    pass


class PinkNoise(WhiteNoise):
    pass


class ClipNoise(WhiteNoise):
    pass


class GrayNoise(WhiteNoise):
    pass


# NoahNoise, commented ugen.


class Crackle(ugn.UGen):
    @classmethod
    def ar(cls, chaos_param=1.5):
        return cls._multi_new('audio', chaos_param)

    @classmethod
    def kr(cls, chaos_param=1.5):
        return cls._multi_new('control', chaos_param)


class Logistic(ugn.UGen):
    @classmethod
    def ar(cls, chaos_param=3.0, freq=1000.0, init=0.5):
        return cls._multi_new('audio', chaos_param, freq, init)

    @classmethod
    def kr(cls, chaos_param=3.0, freq=1000.0, init=0.5):
        return cls._multi_new('control', chaos_param, freq, init)


# Rossler, commented ugen.


class LFNoise0(ugn.UGen):
    @classmethod
    def ar(cls, freq=500.0):
        return cls._multi_new('audio', freq)

    @classmethod
    def kr(cls, freq=500.0):
        return cls._multi_new('control', freq)


class LFNoise1(LFNoise0):
    pass


class LFNoise2(LFNoise0):
    pass


class LFClipNoise(LFNoise0):
    pass


class LFDNoise0(LFNoise0):
    pass


class LFDNoise1(LFNoise0):
    pass


class LFDNoise3(LFNoise0):
    pass


class LFDClipNoise(LFNoise0):
    pass


class Hasher(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0):
        return cls._multi_new('audio', input)

    @classmethod
    def kr(cls, input=0.0):
        return cls._multi_new('control', input)

    def _check_inputs(self):
        if self.rate == 'audio':
            return self._check_sr_as_first_input()
        else:
            return self._check_valid_inputs()


class MantissaMask(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, bits=3):
        return cls._multi_new('audio', input, bits)

    @classmethod
    def kr(cls, input=0.0, bits=3):
        return cls._multi_new('control', input, bits)


class Dust(ugn.UGen):
    @classmethod
    def ar(cls, density=0.0):
        return cls._multi_new('audio', density)

    @classmethod
    def kr(cls, density=0.0):
        return cls._multi_new('control', density)

    @classmethod
    def signal_range(cls):  # override
        return 'unipolar'


class Dust2(ugn.UGen):
    @classmethod
    def ar(cls, density=0.0):
        return cls._multi_new('audio', density)

    @classmethod
    def kr(cls, density=0.0):
        return cls._multi_new('control', density)



"""Demand.sc"""

from ...base import utils as utl
from ...base import builtins as bi
from .. import ugen as ugn
from .. import _graphparam as gpp
from . import line as lne
from . import bufio as bio


class Demand(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, trig, reset, demand_ugens):
        return cls._multi_new('audio', trig, reset, *utl.as_list(demand_ugens))

    @classmethod
    def kr(cls, trig, reset, demand_ugens):
        return cls._multi_new('control', trig, reset, *utl.as_list(demand_ugens))

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(len(self.inputs) - 2, self.rate)


class Duty(ugn.UGen):
    @classmethod
    def ar(cls, dur=1.0, reset=0.0, level=1.0, done_action=0):
        return cls._multi_new('audio', dur, reset, done_action, level)

    @classmethod
    def kr(cls, dur=1.0, reset=0.0, level=1.0, done_action=0):
        return cls._multi_new('control', dur, reset, done_action, level)

    def _check_inputs(self):
        if gpp.ugen_param(self.inputs[0])._as_ugen_rate() == 'demand':
            reset_rate = gpp.ugen_param(self.inputs[1])._as_ugen_rate()
            if reset_rate != 'demand'\
            and reset_rate != 'scalar'\
            and reset_rate != self.rate:
                return (f"reset input is not '{self.rate}' rate: "
                        f"{self.inputs[1]} is '{reset_rate}' rate")
            else:
                return None
        else:
            return self._check_valid_inputs()


class TDuty(Duty):
    @classmethod
    def ar(cls, dur=1.0, reset=0.0, level=1.0, done_action=0, gap_first=0):
        return cls._multi_new('audio', dur, reset, done_action,
                              level, gap_first)

    @classmethod
    def kr(cls, dur=1.0, reset=0.0, level=1.0, done_action=0, gap_first=0):
        return cls._multi_new('control', dur, reset, done_action,
                              level, gap_first)


class DemandEnvGen(ugn.UGen):
    @classmethod
    def ar(cls, level, dur, shape=1, curve=0, gate=1.0, reset=1.0,
           level_scale=1.0, level_bias=0.0, time_scale=1.0, done_action=0):
        gate_rate = gpp.ugen_param(gate)._as_ugen_rate()
        reset_rate = gpp.ugen_param(reset)._as_ugen_rate()
        if gate_rate == 'audio' or reset_rate == 'audio':
            if gate_rate != 'audio':
                gate = lne.K2A.ar(gate)
            if reset_rate != 'audio':
                reset = lne.K2A.ar(reset)
        return cls._multi_new('audio', level, dur, shape, curve, gate, reset,
                              level_scale, level_bias, time_scale, done_action)

    @classmethod
    def kr(cls, level, dur, shape=1, curve=0, gate=1.0, reset=1.0,
           level_scale=1.0, level_bias=0.0, time_scale=1.0, done_action=0):
        return cls._multi_new('control', level, dur, shape, curve, gate, reset,
                              level_scale, level_bias, time_scale, done_action)


class DUGen(ugn.UGen):
    _default_rate = 'demand'

    # // Some n-ary op special cases.

    def linlin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return ((self.prune(inmin, inmax, clip) - inmin) /
                (inmax-inmin) * (outmax - outmin) + outMin)

    def linexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return (bi.pow(outmax / outmin, (self - inmin) / (inmax - inmin)) *
                outmin).prune(outmin, outmax, clip)

    def explin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return (bi.log(self.prune(inmin, inmax, clip) / inmin) /
                bi.log(inmax / inmin) * (outmax - outmin) + outmin)

    def expexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return bi.pow(outmax / outmin,
                      bi.log(self.prune(inmin, inmax, clip / inmin) /
                             bi.log(inmax / inmin)) * outmin)


class Dseries(DUGen):
    @classmethod
    def dr(cls, start=1, step=1, length=float('inf')):
        return cls._multi_new('demand', length, start, step)


class Dgeom(DUGen):
    @classmethod
    def dr(cls, start=1, grow=1, length=float('inf')):
        return cls._multi_new('demand', length, start, grow)


class Dbufrd(DUGen):
    @classmethod
    def dr(cls, bufnum=0, phase=0.0, loop=1.0):
        return cls._multi_new('demand', bufnum, phase, loop)


class Dbufwr(DUGen):
    @classmethod
    def dr(cls, input=0.0, bufnum=0, phase=0.0, loop=1.0):
        return cls._multi_new('demand', bufnum, phase, input, loop)


class ListDUGen(DUGen):
    @classmethod
    def dr(cls, lst, repeats=1):
        return cls._multi_new('demand', repeats, *utl.as_list(lst))


class Dseq(ListDUGen):
    pass


class Dser(ListDUGen):
    pass


class Dshuf(ListDUGen):
    pass


class Drand(ListDUGen):
    pass


class Dxrand(ListDUGen):
    pass


class Dwrand(DUGen):
    @classmethod
    def dr(cls, lst, weights, repeats=1):
        size = len(lst)
        weights = utl.extend(weights, size, 0.0)
        return cls._multi_new('demand', repeats, size, *weights, *lst)


class Dswitch1(DUGen):
    @classmethod
    def dr(cls, lst, index):
        return cls._multi_new('demand', index, *lst)


class Dswitch(Dswitch1):
    pass


class Dwhite(DUGen):
    @classmethod
    def dr(cls, lo=0.0, hi=1.0, length=float('inf')):
        return cls._multi_new('demand', length, lo, hi)


class Diwhite(Dwhite):
    pass


class Dbrown(DUGen):
    @classmethod
    def dr(cls, lo=0.0, hi=1.0, step=0.01, length=float('inf')):
        return cls._multi_new('demand', length, lo, hi, step)


class Dibrown(Dbrown):
    pass


class Dstutter(DUGen):
    @classmethod
    def dr(cls, n, input):
        return cls._multi_new('demand', n, input)


class Dconst(DUGen):
    @classmethod
    def dr(cls, sum, input, tolerance=0.001):
        return cls._multi_new('demand', sum, input, tolerance)


class Dreset(DUGen):
    @classmethod
    def dr(cls, input, reset=0.0):
        return cls._multi_new('demand', input, reset)


class Dpoll(DUGen):
    @classmethod
    def dr(cls, input, label=None, run=1, trig_id=-1):
        return cls._multi_new('demand', input, label, run, trig_id)

    @classmethod
    def _new1(cls, rate, input, label, run, trig_id):  # override
        label = label or f'DemandUGen({type(input).__name__})'
        label = [int(x) for x in bytes(label, 'utf-8')]  # *** TODO: sc ascii method, sclang uses signed values, see Poll.
        obj = cls._create_ugen_object(rate)
        obj._add_to_synth()
        return obj._init_ugen(input, trig_id, run, len(label), *label)


class Dunique(ugn.PseudoUGen):
    _default_rate = 'demand'

    @classmethod
    def dr(cls, source, max_buffer_size=1024, protected=True):
        obj = cls._create_ugen_object('demand')
        obj._init(source, max_buffer_size, protected)
        return obj

    def _init(self, source, max_buffer_size, protected):
        self._protected = bool(protected)
        self._buffer = bio.LocalBuf.new(max_buffer_size)
        self._buffer.clear()
        if self._protected:
            self._iwr = bio.LocalBuf.new(1)
            self._iwr.clear()
            # // Here we also limit to the largest integer a 32bit float
            # // can correctly address.
            self._writer = Dbufwr.new(
                source,  self._buffer,
                Dbufwr.new(Dseries.new(0, 1, 16777216), self._iwr), 1)
        else:
            self._iwr = None
            # // Here we can simply loop.
            self._writer = Dbufwr.new(
                source, self._buffer,
                Dseq.new([Dseries.new(0, 1, max_buffer_size)], float('inf')), 0)

    def _as_ugen_input(self, *_):
        # // We call the writer from each reader.
        brd = self._buffer.first_arg(self._writer)
        if self._protected:
            ird = bio.LocalBuf.new(1)
            ird.clear()
            index = Dbufwr.new(Dseries.new(0, 1, float('inf')), ird)
            overrun = (Dbufrd.new(self._iwr) - Dbufrd.new(ird)) >\
                      self._buffer.num_frames
            # // Catch buffer overrun by switching to a zero length series.
            brd = Dswitch1.new([brd, Dseries.new(length=0)], overrun)
        else:
            index = Dseq.new(
                [Dseries.new(0, 1, self._buffer.num_frames)], float('inf'))
        return Dbufrd.new(brd, index, 1)



"""From Patterns.sc"""

from ...base import builtins as bi
from ...base import stream as stm
from .. import pattern as ptt


### Math patterns ###


class Pseries(ptt.Pattern):
    # // Arithmetic series.
    def __init__(self, start=0.0, step=1.0, length=float('inf')):
        self.start = start
        self.step = step
        self.length = length

    def __embed__(self, inval):
        cur = self.start  # value makes the pattern object to keep state
        length = self.length  # if the parameter is a stream.
        step_stream = stm.stream(self.step)
        outval = stepval = None
        try:
            for _ in bi.counter(length):
                stepval = step_stream.next(inval)
                outval = cur
                cur += stepval
                inval = yield outval
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Pgeom(ptt.Pattern):
    # // Geometric series.
    def __init__(self, start=1.0, grow=1.0, length=float('inf')):
        self.start = start
        self.grow = grow
        self.length = length

    def __embed__(self, inval):
        cur = self.start
        length = self.length
        grow_stream = stm.stream(self.grow)
        outval = growval = None
        try:
            for _ in bi.counter(length):
                growval = grow_stream.next(inval)
                outval = cur
                cur *= growval
                inval = yield outval
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Pbrown(ptt.Pattern):
    def __init__(self, lo=0.0, hi=1.0, step=0.125, length=float('inf')):
        self.lo = lo
        self.hi = hi
        self.step = step
        self.length = length

    def __embed__(self, inval):
        lo_stream = stm.stream(self.lo)
        hi_stream = stm.stream(self.hi)
        step_stream = stm.stream(self.step)
        try:
            loval = lo_stream.next(inval)
            hival = hi_stream.next(inval)
            stepval = step_stream.next(inval)
            current = bi.rrand(loval, hival)
            for _ in bi.counter(self.length):
                loval = lo_stream.next(inval)
                hival = hi_stream.next(inval)
                stepval = step_stream.next(inval)
                current = bi.fold(
                    self._calc_next(current, stepval), loval, hival)
                inval = yield current
        except stm.StopStream:
            pass
        return inval

    def _calc_next(self, current, step):
        return current + bi.xrand2(step)

    # storeArgs


class Pgbrown(Pbrown):
    def _calc_next(self, current, step):
        return current * (1 + bi.xrand2(step))


class Pwhite(ptt.Pattern):
    def __init__(self, lo=0.0, hi=1.0, length=float('inf')):
        self.lo = lo
        self.hi = hi
        self.length = length

    def __embed__(self, inval):
        lo_stream = stm.stream(self.lo)
        hi_stream = stm.stream(self.hi)
        hival = loval = None
        try:
            for _ in bi.counter(self.length):
                loval = lo_stream.next(inval)
                hival = hi_stream.next(inval)
                inval = yield bi.rrand(loval, hival)
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Pprob(ptt.Pattern):
    def __init__(self, distribution, lo=0.0, hi=1.0,
                 table_size=None, length=float('inf')):
        self.distribution = distribution
        self.lo = lo
        self.hi = hi
        # Patterns arguments should be constant after instantiation.
        self.table_size = table_size or max(64, len(distribution))
        self.table = bi.as_random_table(distribution, self.table_size)
        self.length = length

    def __embed__(self, inval):
        table = self.table
        lo_stream = stm.stream(self.lo)
        hi_stream = stm.stream(self.hi)
        lval = hval = None
        try:
            for _ in bi.counter(self.length):
                lval = lo_stream.next(inval)
                hval = hi_stream.next(inval)
                inval = yield (bi.table_rand(table) * (hval - lval)) + lval
        except stm.StopStream:
            pass
        return inval


class Pproduct(ptt.Pattern):  # Was PstepNfunc.
    def __init__(self, patterns, func=None):
        self.patterns = patterns
        self.func = func or (lambda values: values)

    def __embed__(self, inval):
        # If there wasn't inval things would be much easier.
        # for t in itertools.product(*self.patterns): ...
        patterns = self.patterns
        size = len(patterns)
        max_level = size - 1
        streams = [None] * size
        values = [None] * size
        yield from self._recgen(inval, 0, max_level, patterns, streams, values)

    def _recgen(self, inval, level, max_level, patterns, streams, values):
        try:
            streams[level] = stm.stream(patterns[level])
            while True:
                values[level] = streams[level].next(inval)
                if level < max_level:
                    yield from self._recgen(
                        inval, level + 1, max_level, patterns, streams, values)
                else:
                    yield self.func(values)
        except stm.StopStream:
            pass
        return inval

    # storeArgs


# Superseded by PstepNfunc (Pproduct).
# class Pstep2add(ptt.Pattern)
# class Pstep3add(ptt.Pattern)
# class PstepNadd(PstepNfunc)



"""Platform.sc"""

from pathlib import Path
import tempfile
import os
import sys
# import site
import subprocess
import threading

from . import classlibrary as clb


clb.ClassLibrary.late_imports(__name__, ('sc3.base.main', '_libsc3'))


__all__ = ['Platform']


class MetaPlatform(type):
    @property
    def name(cls):
        return sys.platform

    @property
    def home_dir(cls):
        return _libsc3.main.platform.home_dir

    @property
    def support_dir(cls):
        return _libsc3.main.platform.support_dir

    @property
    def config_dir(cls):
        return _libsc3.main.platform.config_dir

    @property
    def resource_dir(cls):
        return _libsc3.main.platform.resource_dir

    @property
    def synthdef_dir(cls):
        return _libsc3.main.platform.synthdef_dir

    @property
    def recording_dir(cls):
        return _libsc3.main.platform.recording_dir

    @property
    def tmp_dir(cls):
        return _libsc3.main.platform.tmp_dir

    @property
    def installation_dir(cls):
        return _libsc3.main.platform.installation_dir

    @installation_dir.setter
    def installation_dir(cls, value):
        _libsc3.main.platform.installation_dir = value

    @property
    def bin_dir(cls):
        return _libsc3.main.platform.bin_dir

    @property
    def default_server_cmd(cls):
        return _libsc3.main.platform.default_server_cmd

    @property
    def SCSYNTH_CMD(cls):
        return _libsc3.main.platform.SCSYNTH_CMD

    @property
    def SUPERNOVA_CMD(cls):
        return _libsc3.main.platform.SUPERNOVA_CMD

    @staticmethod
    def _cmd_line(*popenargs, sync=False):  # sclang unixCmd no pid return.
        if sync:
            subprocess.call(*popenargs)
        else:
            run = lambda: subprocess.call(*popenargs)
            threading.Thread(target=run, daemon=True).start()

    def killall(cls, program_name):
        '''
        Kill a server process by name. Utility method used to stop a lost
        server in the local machine, e.g. when is not possible to reboot or
        connect and the process is retaining the port.
        On Unix this method executes ``killall -SIGKILL program_name``.
        On Windows it executes ``taskkill /F /IM program_name``.
        '''
        return _libsc3.main.platform.killall(program_name)


class Platform(metaclass=MetaPlatform):
    def _startup(self):
        pass

    def _shutdown(self):
        pass

    @property
    def name(self):
        '''The name of the platform as returned by `sys.platform`.'''
        return sys.platform

    @property
    def home_dir(self):  # Was userHomeDir.
        '''User home directory.'''
        return Path.home()

    @property
    def support_dir(self):  # Was userAppSupportDir.
        '''The directory containing SuperCollider user resources.'''
        raise NotImplementedError

    @property
    def config_dir(self):  # Was userConfigDir.
        '''The directory containing SuperCollider user configuration.'''
        raise NotImplementedError

    @property
    def resource_dir(self):  # userAppSupportDir.
        '''The directory containing SuperCollider installation resources.'''
        raise NotImplementedError

    @property
    def synthdef_dir(self):
        '''The directory where synthdef files are sotred by default.'''
        raise NotImplementedError

    @property
    def recording_dir(self):
        '''The directory where recording files are sotred by default.'''
        raise NotImplementedError

    @property
    def tmp_dir(self):
        '''The temporal directory, set to system default `tmp` directory.'''
        return Path(tempfile.gettempdir())

    @property
    def installation_dir(self):
        '''Installation directory, in OSX it's the app bundle.'''
        return Path(self._installation_dir)

    @installation_dir.setter
    def installation_dir(self, value):
        # Attribute _installation_dir and os.environ['PATH'] mut be
        # properly initialized by _startup().
        old_path_dir = str(self.bin_dir)
        self._installation_dir = str(value)
        os.environ['PATH'] = os.environ['PATH'].replace(
            old_path_dir, str(self.bin_dir))

    @property
    def bin_dir(self):
        '''The directory where the server programs are located.'''
        raise NotImplementedError


class UnixPlatform(Platform):
    SCSYNTH_CMD = 'scsynth'
    SUPERNOVA_CMD = 'supernova'

    def killall(self, program_name):
        cmd = ['killall', '-9', program_name]
        type(self)._cmd_line(cmd)


class LinuxPlatform(UnixPlatform):
    def _startup(self):
        # // default jack port hookup
        os.environ['SC_JACK_DEFAULT_INPUTS'] = 'system'
        os.environ['SC_JACK_DEFAULT_OUTPUTS'] = 'system'
        # // automatically start jack when booting the server
        # // can still be overridden with JACK_NO_START_SERVER
        os.environ['JACK_START_SERVER'] = 'true'

        self._installation_dir = '/usr'  # whereis scsynth ../
        os.environ['PATH'] += os.pathsep + str(self.bin_dir)
        self.default_server_cmd = self.SCSYNTH_CMD
        # // load user startup file
        # self.load_startup_file()

    def _shutdown(self):
        pass

    @property
    def support_dir(self):
        return self.home_dir / Path('.local/share/SuperCollider')

    @property
    def config_dir(self):
        return self.home_dir / Path('.config/SuperCollider')

    @property
    def resource_dir(self):
        return self.installation_dir / Path('share/SuperCollider')

    @property
    def synthdef_dir(self):
        return self.support_dir / Path('synthdefs')

    @property
    def recording_dir(self):
        return self.support_dir / Path('Recordings')

    @property
    def bin_dir(self):
        return self.installation_dir / Path('bin')


class DarwinPlatform(UnixPlatform):
    def _startup(self):
        self._installation_dir = '/Applications/SuperCollider.app'
        os.environ['PATH'] += os.pathsep + str(self.bin_dir)
        self.default_server_cmd = self.SCSYNTH_CMD

    @property
    def support_dir(self):
        return self.home_dir / Path('Library/Application Support/SuperCollider')

    @property
    def config_dir(self):
        return self.home_dir / Path('Library/Application Support/SuperCollider')

    @property
    def resource_dir(self):
        return self.installation_dir / Path('Contents/Resources')

    @property
    def synthdef_dir(self):
        return self.support_dir / Path('synthdefs')

    @property
    def recording_dir(self):
        return self.home / Path('Music/SuperCollider Recordings')

    @property
    def bin_dir(self):
        return self.resource_dir


class WindowsPlatform(Platform):
    SCSYNTH_CMD = 'scsynth.exe'
    SUPERNOVA_CMD = 'supernova.exe'

    def killall(self, program_name):
        cmd = ['taskkill', '/F', '/IM', program_name]
        type(self)._cmd_line(cmd)


class Win32Platform(WindowsPlatform):
    def _startup(self):
        from . import _knownpaths as kp

        self._local_app_data = kp.get_path(kp.FOLDERID.LocalAppData, 0)
        self._documents = kp.get_path(kp.FOLDERID.Documents, 0)

        program_files_x86 = kp.get_path(kp.FOLDERID.ProgramFilesX86, 0)
        program_files_x64 = kp.get_path(kp.FOLDERID.ProgramFilesX64, 0)
        folders = list(Path(program_files_x86).glob('SuperCollider*'))
        folders += list(Path(program_files_x64).glob('SuperCollider*'))

        self._installation_dir = str(self._get_lastest_version(folders))
        os.environ['PATH'] += str(self.bin_dir) + os.pathsep
        self.default_server_cmd = self.SCSYNTH_CMD
        # // load user startup file
        # self.load_startup_file()

    def _shutdown(self):
        pass

    @staticmethod
    def _get_lastest_version(path_list):
        # Get the folder with the lastest version.
        # Folder name format sould be 'SuperCollider-X.X.X'.
        res = []
        for path in path_list:
            name = path.name
            version = name.split('-')[1]
            version = version.split('.')
            version = [int(n) for n in version if n.isdigit()]
            res.append((version, path))
        return sorted(res, key=lambda x: x[0])[-1][1]

    @property
    def support_dir(self):
        return Path(self._local_app_data) / Path('SuperCollider')

    @property
    def config_dir(self):
        return self.support_dir()

    @property
    def resource_dir(self):
        return self.installation_dir

    @property
    def synthdef_dir(self):
        return self.support_dir / Path('synthdefs')

    @property
    def recording_dir(self):
        return Path(self._documents) / Path('Recordings')

    @property
    def bin_dir(self):
        return self.installation_dir


class CygwinPlatform(WindowsPlatform):
    ...



"""Operand.sc"""

from . import absobject as aob


__all__ = ['Operand']


class Operand(aob.AbstractObject):
    def __init__(self, value=None):
        if isinstance(value, Operand):
            self.value = value.value
        else:
            self.value = value

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        self.__value = value


    ### AbstractObject interface ###

    def _compose_unop(self, selector):
        return type(self)(selector(self.value))

    def _compose_binop(self, selector, other):
        a = self.value
        b = other.value if isinstance(other, Operand) else other
        return type(self)(selector(a, b))

    def _rcompose_binop(self, selector, other):
        a = other.value if isinstance(other, Operand) else other
        b = self.value
        return type(self)(selector(a, b))

    def _compose_narop(self, selector, *args):
        return type(self)(selector(self.value, *args))


    def __hash__(self):
        return self.value.__hash__()

    def __eq__(self, value):
        if isinstance(value, Operand):
            return self.value.__eq__(value.value)
        else:
            return self.value.__eq__(value)

    def __repr__(self):
        return f'{type(self).__name__}({repr(self.value)})'


    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __str__(self):
        return str(self.value)

    # printOn
    # storeOn



"""NodeWatcher.sc"""

from ..base import responsedefs as rdf
from ..base import systemactions as sac
from ..base import model as mdl


# // Watches a server address for node-related messages


class NodeWatcher():
    # // Watches registered nodes and sets their isPlaying/isRunning flag.
    # // A node needs to be registered to be addressed, other nodes are ignored.

    def __init__(self, server):
        self._server = server
        self._nodes = dict()
        self._responders = []
        self._is_watching = False
        for cmd in self.cmds:
            method = '_' + cmd[1:]
            osc_func = rdf.OscFunc(
                (lambda mthd: lambda msg, *_: self.respond(mthd, msg))(method),
                cmd, self._server.addr)
            osc_func.permanent = True
            # osc_func.disable()  # *** NOTE: no sense if self.start() is not manual.
            self._responders.append(osc_func)
        sac.CmdPeriod.add(self.__on_cmd_period)
        sac.ServerBoot.add(self._server, self.__on_server_boot)
        sac.ServerQuit.add(self._server, self.__on_server_quit)
        self.start()  # *** NOTE: could be manual: server._node_watcher.start().

    @property
    def server(self):
        return self._server

    @property
    def nodes(self):
        return self._nodes

    @property
    def responders(self):
        return self._responders

    @property
    def is_watching(self):
        return self._is_watching

    @property
    def cmds(self):
        return ('/n_go', '/n_end', '/n_off', '/n_on', '/n_move', '/n_info')

    def respond(self, method, msg):
        node = self._nodes.get(msg[1])
        if node is not None:
            getattr(self, method)(node)  # *** NOTE: podría pasar también msg completo.

    def start(self):
        if not self._is_watching:
            for item in self._responders:
                item.enable()
                self._is_watching = True

    def stop(self):
        if self._is_watching:
            for item in self._responders:
                item.free()
            self._is_watching = False
            # self.free()  # *** BUG: sclang call Object.free that does nothing.

    def register(self, node, playing=True, running=True):
        if not self._server._status_watcher.server_running:
            # self._nodes.clear()  # *** BUG: sclang usa removeAll sin argumentos, que no hace nada!
            return
        if self._is_watching:
            if self._nodes.get(node.node_id) is None:
                if playing: node._is_playing = True
                if running: node._is_running = True
            self._nodes[node.node_id] = node

    def unregister(self, node):
        self._nodes.pop(node.node_id, None)

    def registered(self, node):  # Was node_is_playing, better not to use is_*.
        return node.node_id in self._nodes

    def _clear(self):
        # // we must copy 'nodes' b/c a /n_end dependant (NotificationCenter)
        # // might add or remove nodes from the collection
        # // NEVER iterate over a collection that might change
        for node in self._nodes.copy().values():
            node._is_playing = None
            node._is_running = None
            mdl.NotificationCenter.notify(node, '/n_end')
        self._nodes = dict()

    def _n_go(self, node):
        node._is_playing = True
        node._is_running = True
        mdl.NotificationCenter.notify(node, '/n_go')

    def _n_end(self, node):
        self.unregister(node)
        node._is_playing = False
        node._is_running = False
        mdl.NotificationCenter.notify(node, '/n_end')

    def _n_off(self, node):  # response to /n_run 0
        node._is_running = False
        mdl.NotificationCenter.notify(node, '/n_off')

    def _n_on(self, node):  # response to /n_run 1
        node._is_running = True
        mdl.NotificationCenter.notify(node, '/n_on')

    def _n_move(self, node):  # response to /n_before, /n_after, /n_order
        mdl.NotificationCenter.notify(node, '/n_move')

    def _n_info(self, node):  # response to /n_query
        mdl.NotificationCenter.notify(node, '/n_info')


    ### System Actions ###

    def __on_cmd_period(self):
        self._clear()

    def __on_server_boot(self, _):
        self._clear()

    def __on_server_quit(self, _):
        self._clear()
